unit DosSock;

interface

Type
    TProtocolStruct = record
    	Domain, SocketType, Protocol: LongInt;
        FNSocket: Function(Domain, SocketType, Protocol: Longint): Pointer;
        FNClose: Function(Sock: Pointer): LongInt;
        FNGetLastError: Function: LongInt;
        FNSend: Function(Sock: Pointer; Var Addr; AddrLen, Flags: Longint): Longint;
        FNRecv: Function(Sock: Pointer; Var Addr; AddrLen, Flags: Longint): Longint;
        FNBind: Function(Sock: Pointer; Var Addr; AddrLen: Longint): Boolean;
        FNListen: Function(Sock: Pointer; MaxConnect: Longint): Boolean;
        FNAccept: Function(Sock: Pointer; Var Addr; Var Addrlen: Longint): Pointer;
        FNConnect: Function(Sock: Pointer; Var Addr; Addrlen: Longint): Boolean;
        FNShutdown: Function(Sock: Pointer; How: Longint): Longint;
        FNGetSocketName: Function(Sock: Pointer; Var Addr; Var Addrlen: Longint): Longint;
        FNGetPeerName: Function(Sock: Pointer; Var Addr; Var Addrlen: Longint): Longint;
        FNSetSocketOptions: Function(Sock: Pointer; Level, OptName: Longint; Var OptVal; OptLen: Longint): Longint;
        FNGetSocketOptions: Function(Sock: Pointer; Level, OptName: Longint; Var OptVal; Var OptLen: Longint): Longint;
        FNioctlsocket: Function(Sock: Pointer; Option: LongInt; d: Pointer): LongInt;
    end;


Function socket(Domain, SocketType, Protocol: Longint): Longint;
Function CloseSocket(Sock: LongInt): Longint;
Function GetLastError: LongInt;
Function Send(Sock: Longint; Var Addr; AddrLen, Flags: Longint): Longint;
Function Recv(Sock: Longint; Var Addr; AddrLen, Flags: Longint): Longint;
Function Bind(Sock: Longint; Var Addr; AddrLen: Longint): Boolean;
Function Listen(Sock, MaxConnect: Longint): Boolean;
Function Accept(Sock: Longint; Var Addr; Var Addrlen: Longint): Longint;
Function Connect(Sock: Longint; Var Addr; Addrlen: Longint): Boolean;
Function Shutdown(Sock: Longint; How: Longint): Longint;
Function GetSocketName(Sock: Longint; Var Addr; Var Addrlen: Longint): Longint;
Function GetPeerName(Sock: Longint; Var Addr; Var Addrlen: Longint): Longint;
Function SetSocketOptions(Sock, Level, OptName: Longint; Var OptVal; OptLen: Longint): Longint;
Function GetSocketOptions(Sock, Level, OptName: Longint; Var OptVal; Var OptLen: Longint): Longint;
Function ioctlsocket(Sock: LongInt; Option: LongInt; d: Pointer): LongInt;
Function AddProtocol(ProtocolStruct: TProtocolStruct): Boolean;

implementation

Uses
    Sockets;

Type
	PSockStruct = ^TSockStruct;
    TSockStruct = record
        Ptr: Pointer;
        ProtocolIndex: Integer;
    end;

Var
	LastError: LongInt;
    Protocols: Array[1..100] of TProtocolStruct;


Function socket(Domain, SocketType, Protocol: Longint): Longint;
Var
    Ptr: Pointer;
    i, PIndex: Integer;
	PSock: PSockStruct;
begin
	Socket := INVALID_SOCKET;
    { Decide what protocol to use based on Domain, Socket Type and }
    { Protocol.                                                    }
    PIndex := -1;
    for i := Low(Protocols) to High(Protocols) do begin
        if (Protocols[i].Domain = Domain)
          and (Protocols[i].SocketType = SocketType)
          and (Protocols[i].Protocol = Protocol) then begin
            PIndex := i;
            Break;
        end;
    end;
    { If there's no protocol registered to satisfy the user's request, }
    { we exit with WSAEPROTONOSUPPORT.                                 }
    if PIndex = -1 then begin
        LastError := WSAEPROTONOSUPPORT;
        Exit;
    end;
    { Request the selected protocol to create a new socket             }
    Ptr := Protocols[PIndex].FNSocket(Domain, SocketType, Protocol);
    if (Ptr = nil) then begin
        LastError := Protocols[PIndex].FNGetLastError();
        Exit;
    end;
    { We've found a protocol to meet the user's request. Now it's time }
    { to allocate a handle for the new socket                          }
    New(PSock);
    PSock^.Ptr := Ptr;
    PSock^.ProtocolIndex := PIndex;
    Socket := LongInt(PSock);
end;

Function CloseSocket(Sock: LongInt): Longint;
Var
    PSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        CloseSocket := SOCKET_ERROR;
        LastError   := WSAENOTSOCK;
    end else begin
	    PSock := PSockStruct(Sock);
    	CloseSocket := Protocols[PSock^.ProtocolIndex].FNClose(PSock^.Ptr);
        LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
        Dispose(PSock);
    end;
end;

Function GetLastError: LongInt;
begin
	GetLastError := LastError;
end;

Function Send(Sock: Longint; Var Addr; AddrLen, Flags: Longint): Longint;
Var
    PSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        Send := SOCKET_ERROR;
        LastError   := WSAENOTSOCK;
    end else begin
	    PSock := PSockStruct(Sock);
    	Send := Protocols[PSock^.ProtocolIndex].FNSend(PSock^.Ptr,
	      Addr, AddrLen, Flags);
        LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
    end;
end;

Function Recv(Sock: Longint; Var Addr; AddrLen, Flags: Longint): Longint;
Var
    PSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        Recv := SOCKET_ERROR;
        LastError := WSAENOTSOCK;
    end else begin
	    PSock := PSockStruct(Sock);
    	Recv := Protocols[PSock^.ProtocolIndex].FNRecv(PSock^.Ptr,
	      Addr, AddrLen, Flags);
        LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
    end;
end;

Function Bind(Sock: Longint; Var Addr; AddrLen: Longint): Boolean;
Var
    PSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        Bind := False;
        LastError   := WSAENOTSOCK;
    end else begin
	    PSock := PSockStruct(Sock);
    	Bind := Protocols[PSock^.ProtocolIndex].FNBind(PSock^.Ptr,
	      Addr, AddrLen);
        LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
    end;
end;

Function Listen(Sock, MaxConnect: Longint): Boolean;
Var
    PSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        Listen := False;
        LastError   := WSAENOTSOCK;
    end else begin
	    PSock := PSockStruct(Sock);
    	Listen := Protocols[PSock^.ProtocolIndex].FNListen(PSock^.Ptr,
	      MaxConnect);
        LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
    end;
end;

Function Accept(Sock: Longint; Var Addr; Var Addrlen: Longint): Longint;
Var
	Ptr: Pointer;
    PSock, NewSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        Accept := SOCKET_ERROR;
        LastError   := WSAENOTSOCK;
        Exit;
    end;
    PSock := PSockStruct(Sock);
    Ptr := Protocols[PSock^.ProtocolIndex].FNAccept(PSock^.Ptr,
      Addr, AddrLen);
    if Ptr = nil then begin
        Accept := SOCKET_ERROR;
    end else begin
        New(NewSock);
        NewSock^.Ptr := Ptr;
        NewSock^.ProtocolIndex := PSock^.ProtocolIndex;
        Accept := LongInt(NewSock);
    end;
    LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
end;

Function Connect(Sock: Longint; Var Addr; Addrlen: Longint): Boolean;
Var
    PSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        Connect := False;
        LastError   := WSAENOTSOCK;
    end else begin
	    PSock := PSockStruct(Sock);
    	Connect := Protocols[PSock^.ProtocolIndex].FNConnect(PSock^.Ptr,
	      Addr, AddrLen);
        LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
    end;
end;

Function Shutdown(Sock: Longint; How: Longint): Longint;
Var
    PSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        Shutdown := SOCKET_ERROR;
        LastError   := WSAENOTSOCK;
    end else begin
	    PSock := PSockStruct(Sock);
    	Shutdown := Protocols[PSock^.ProtocolIndex].FNShutdown(PSock^.Ptr,
	      How);
        LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
    end;
end;

Function GetSocketName(Sock: Longint; Var Addr; Var Addrlen: Longint): Longint;
Var
    PSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        GetSocketName := SOCKET_ERROR;
        LastError   := WSAENOTSOCK;
    end else begin
	    PSock := PSockStruct(Sock);
    	GetSocketName := Protocols[PSock^.ProtocolIndex].FNGetSocketName(PSock^.Ptr,
	      Addr, AddrLen);
        LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
    end;
end;

Function GetPeerName(Sock: Longint; Var Addr; Var Addrlen: Longint): Longint;
Var
    PSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        GetPeerName := SOCKET_ERROR;
        LastError   := WSAENOTSOCK;
    end else begin
	    PSock := PSockStruct(Sock);
    	GetPeerName := Protocols[PSock^.ProtocolIndex].FNGetPeerName(PSock^.Ptr,
	      Addr, AddrLen);
        LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
    end;
end;

Function SetSocketOptions(Sock, Level, OptName: Longint; Var OptVal; OptLen: Longint): Longint;
Var
    PSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        SetSocketOptions := SOCKET_ERROR;
        LastError   := WSAENOTSOCK;
    end else begin
	    PSock := PSockStruct(Sock);
    	SetSocketOptions := Protocols[PSock^.ProtocolIndex].FNSetSocketOptions(PSock^.Ptr,
	      Level, OptName, OptVal, OptLen);
        LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
    end;
end;

Function GetSocketOptions(Sock, Level, OptName: Longint; Var OptVal; Var OptLen: Longint): Longint;
Var
    PSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        GetSocketOptions := SOCKET_ERROR;
        LastError   := WSAENOTSOCK;
    end else begin
	    PSock := PSockStruct(Sock);
    	GetSocketOptions := Protocols[PSock^.ProtocolIndex].FNGetsocketOptions(PSock^.Ptr,
	      Level, OptName, OptVal, OptLen);
        LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
    end;
end;

Function ioctlsocket(Sock: LongInt; Option: LongInt; d: pointer): LongInt;
Var
    PSock: PSockStruct;
begin
    if (Sock <= 0) then begin
        ioctlsocket := SOCKET_ERROR;
        LastError   := WSAENOTSOCK;
    end else begin
        PSock := PSockStruct(Sock);
    	ioctlsocket := Protocols[PSock^.ProtocolIndex].FNioctlsocket(PSock^.Ptr,
	      Option, d);
        LastError := Protocols[PSock^.ProtocolIndex].FNGetLastError();
    end;
end;

Function AddProtocol(ProtocolStruct: TProtocolStruct): Boolean;
Var
	i: Integer;
begin
    for i := Low(Protocols) to High(Protocols) do begin
        if Protocols[i].Domain < 0 then begin
        	Protocols[i] := ProtocolStruct;
            AddProtocol := True;
            Exit;
        end;
    end;
    AddProtocol := False;
end;


Var
	i: Integer;
initialization
	LastError := 0;
    for i := Low(Protocols) to High(Protocols) do begin
        Protocols[i].Domain := -1;
    end;
end.