Unit PktDrv;

{$X+}
{$MODE Delphi}

Interface

Uses
    Go32, Crt, Objects, SysUtils, Util, Network;


Type
	{ Adapter Object }
	TPacketDriverAdapter = class(TNetworkAdapter)
	private
        FPDInt: BYTE;
        FPDHandle: WORD;
        FProtCount: Integer;
    protected
        procedure CheckStatus; override;
        Function  GetPacket(Var Packet: TEthernetPacket): Boolean; override;
        function  CheckForPD(Int: BYTE): Boolean;
		function  IntAddProtocol(TypeLen: WORD): Boolean; override;
		function  IntRemoveProtocol(TypeLen: WORD): Boolean; override;
    public
        constructor Create;
        Function    GetAddress(var Addr: TEthernetAddress): Boolean; override;
		Function    SendPkt(Buffer: Pointer; BufLen: Word): Boolean; override;
	end;


implementation

Var
    { Temporary 2K DOS Storage }
	GlobalDosTempSeg: WORD;
	GlobalDosTempSel: WORD;
    { Real-Mode Pointer to Real-Mode Receiver routine }
    RMCB_SegInfo: TSegInfo;
    {**** Start of locked Data. DON'T MOVE ****}
    Lock_begin: BYTE;
    iCnt: Integer;
    RMCB_regs: TRealRegs;
    GlobalDosRecvSeg: WORD;
	GlobalDosRecvSel: WORD;
    PacketQueue: Array[1..64] of TEthernetPacket;
    Lock_end: BYTE;
    {**** End of locked data ****}


{******************* Begin of Locked Code ********************}
Procedure PDAlloc;
begin
    RMCB_regs.ES := GlobalDOSRecvSeg;
    RMCB_regs.DI := 0;
end;

Procedure PDCopied;
begin
    { Search for a free buffer }
	for iCnt := Low(PacketQueue) to High(PacketQueue) do begin
        if PacketQueue[iCnt].Header.TypeLen = 0 then begin
            DOSMemGet(GlobalDOSRecvSeg, 0, PacketQueue[iCnt], SizeOf(TEthernetPacket));
            Break;
        end;
    end;
end;

{$ASMMODE INTEL}
Procedure PDReceiverAsm;
begin
    asm
        pusha
        push ds
        push es
        push esi
        push edi
        mov ax,es
        mov ds,ax
    end;
    if RMCB_regs.AX = 0 then PDAlloc else if RMCB_regs.AX = 1 then PDCopied;
    asm
        pop edi
        pop esi
        pop es
        pop ds
        mov eax,[esi]
        mov es:[edi+42],eax
        add word ptr es:[edi+46],4
        popa
        leave
        iret
    end;
end;

procedure PDReceiverEnd; begin; end;
{******************* End of Locked Code ********************}

constructor TPacketDriverAdapter.Create;
begin
    Inherited Create;
    FProtCount := 0;
end;

procedure TPacketDriverAdapter.CheckStatus;
begin
    for FPDInt := $60 to $80 do begin
        if CheckForPD(FPDInt) then begin
            FStatus := asOk;
            Exit;
        end;
    end;
	FStatus := asError;
end;

function TPacketDriverAdapter.CheckForPD(Int: Byte): Boolean;
Const
	Signature = 'PKT DRVR';
Var
	i: Integer;
	SegInfo: TSegInfo;
	Buff: Array[0..20] of char;
begin
	{ Check to see if there's a packet driver installed at Int }
	get_rm_interrupt(Int, SegInfo);
	for i := 0 to 10 do begin
		FillChar(Buff, SizeOf(Buff), 0);
        DOSMemGet(SegInfo.Segment, LongInt(SegInfo.Offset) + i, Buff, Length(Signature));
		if StrPas(Buff) = Signature then begin
			Result := True;
			Exit;
		end;
	end;
	{ No Packet Driver found }
	Result := False;
end;

function TPacketDriverAdapter.IntAddProtocol(TypeLen: WORD): Boolean;
Var
	IfClass: BYTE;
    Regs: Registers;
    Buffer: Array[0..255] of char;
begin
    if FProtCount <> 0 then begin
    	Inc(FProtCount);
        Result := True;
        Exit;
    end;

	Result := False;

	{ Obtain info for the packet driver }
	With Regs do begin
        AH := 1;
        AL := $FF;
    	BX := 0;
    	RealIntr(FPDInt, Regs);
    	if (Regs.Flags and 1) = 1 then Exit;
        IfClass := CH;
        DOSMemGet(DS, SI, Buffer, 40);
        FAdapterName := StrPas(Buffer);
    end;

    { Check for the recovery file. If it exists, we have to close the   }
    { handle first. This recover file gets deleted when the packet      }
    { driver is closed gracefully.                                      }
    FPDHandle := StrToIntDef(FileToString('PktDrv.Hdl'), 0);
    if FPDHandle > 0 then With Regs do begin
		AH := 3;
		BX := FPDHandle;
        RealIntr(FPDInt, Regs);
    end;

    { Register with packet driver }
    TypeLen := 0;
    With Regs do begin
        AH := 2;
        AL := IfClass;
        BX := $FFFF;
        CX := 0;
        DL := 0;
        // TypeLen
        DosMemPut(GlobalDosTempSeg, 0, TypeLen, 2);
        DS := GlobalDosTempSeg;
        SI := 0;
		// Receiver Routine
        ES := RMCB_SegInfo.Segment;
        DI := LongInt(RMCB_SegInfo.Offset);
		// Call Packet Driver
        RealIntr(FPDInt, Regs);
        if (Flags and 1) = 1 then Exit;
		FPDHandle := AX;
    end;

    { Save a recovery file to Disk to resume on App crashes.            }
    { The next time we initialize the stack we will check for the       }
    { existance of this file. (See above)                               }
    StringToFile(IntToStr(FPDInt), 'PktDrv.Hdl');

	{ Set receive mode to 3: receive packets for this board or broadcast }
	with Regs do begin
		AH := 20;
		BX := FPDHandle;
		CX := 3;
		RealIntr(FPDInt, Regs);
	end;
	{ Done! }
    FProtCount := 1;
    Result := True;
end;

function TPacketDriverAdapter.IntRemoveProtocol(TypeLen: WORD): Boolean;
Var
	Regs: Registers;
begin
    Result := True;
    if FProtCount <> 1 then begin
    	Dec(FProtCount);
        Exit;
    end;
    { Unregister with packet driver }
    with Regs do begin
		AH := 3;
		BX := FPDHandle;
		RealIntr(FPDInt, Regs);
    end;
    { Done! }
    DeleteFile('ktDrv.Hdl');
    FProtCount := 0;
end;

Function TPacketDriverAdapter.SendPkt(Buffer: Pointer; BufLen: Word): Boolean;
Var
    Regs: TRealRegs;
begin
    SendPkt := False;
    if FStatus <> asOk then Exit;
    if BufLen < 60 then BufLen := 60;
    DosMemPut(GlobalDosTempSeg, 0, Buffer^, BufLen);
    With Regs do begin
        AH := 4;
        BX := FPDHandle;
        DS := GlobalDosTempSeg;
        ES := GlobalDosTempSeg;
        SI := 0;
        CX := BufLen;
        RealIntr(FPDInt, Regs);
        if (Flags and 1) = 1 then Exit;
    end;
    SendPkt := True;
end;

Function TPacketDriverAdapter.GetAddress(var Addr: TEthernetAddress): Boolean;
Var
    P: DWORD;
    Regs: TRealRegs;
    Segment, Selector: WORD;
begin
    GetAddress := False;
    if (FStatus <> asOk) then Exit;
    P := Global_DOS_Alloc(SizeOf(TEthernetAddress));
    Segment  := Word(P shr 16);
    Selector := Word(P);
    With Regs do begin
    	AH := 6;
        BX := FPDHandle;
        CX := SizeOf(TEthernetAddress);
	    ES := Segment;	
    	DI := 0;		
        RealIntr(FPDInt, Regs);
        if (Flags and 1) = 1 then Exit;
    end;
	DOSMemGet(Segment, 0, Addr, SizeOf(Addr));
    Global_DOS_Free(Selector);
    GetAddress := True;
end;

Function TPacketDriverAdapter.GetPacket(Var Packet: TEthernetPacket): Boolean;
Var
	i: Integer;
begin
    for i := Low(PacketQueue) to High(PacketQueue) do begin
        if PacketQueue[i].Header.TypeLen <> 0 then begin
            Packet := PacketQueue[i];
            PacketQueue[i].Header.TypeLen := 0;
            Result := True;
            Exit;
        end;
    end;
	Result := False;
end;


Var
	P: DWORD;
initialization
    { Clear Packet Queue }
    FillChar(PacketQueue, SizeOf(PacketQueue), 0);
    { Allocate 2K temporary DOS buffer }
    P := Global_DOS_Alloc(2048);
    GlobalDosTempSeg := Word(P shr 16);
	GlobalDosTempSel := Word(P);
    { Allocate 2K temporary DOS buffer for receiving }
    P := Global_DOS_Alloc(2048);
    GlobalDosRecvSeg := Word(P shr 16);
	GlobalDosRecvSel := Word(P);
    // Allocate DOS Callback
    if not lock_code(@PDAlloc, LongInt(@PDReceiverEnd) - LongInt(@PDAlloc))
	  or not lock_data(DOSMEMSELECTOR, sizeof(DOSMEMSELECTOR))
      or not lock_data(Lock_begin, LongInt(@Lock_end) - LongInt(@Lock_begin)) then
      raise exception.Create('Error locking code and/or data for packet driver');
    if not get_rm_callback(@PDReceiverAsm, RMCB_regs, RMCB_SegInfo) then
      raise exception.Create('Error creating real-mode callback for packet driver');
finalization
	// Unlock data & code; remove allocated DOS memory
    free_rm_callback(RMCB_SegInfo);
    unlock_code(@PDAlloc, LongInt(@PDReceiverEnd) - LongInt(@PDAlloc));
	unlock_data(DOSMEMSELECTOR, sizeof(DOSMEMSELECTOR));
    unlock_data(Lock_begin, LongInt(@Lock_end) - LongInt(@Lock_begin));
    Global_DOS_Free(GlobalDosTempSel);
    Global_DOS_Free(GlobalDosRecvSel);
end.