Program Ping;

uses
	Crt, SysUtils, Network, PktDrv, IP, Sockets, Util, DNS;

Const
	MY_IP			= '128.128.1.217';
	MY_NETMASK		= '255.255.255.0';
	MY_GATEWAY		= '128.128.1.100';
	DNS_SERVER		= '128.128.1.100';

Var
	c: Char;
	ToRead: DWORD;
	ServerPort: WORD;
	IPHandle: Integer;
	ServerIP: TIPAddress;
	SockAddr: TInetSockAddr;
	IPConf: TIPAdapterConfig;
	Adapter: TPacketdriverAdapter;
	HSocket, ConnTimeout, i: Integer;
begin

	// Check Params
	try
		if ParamCount <> 2 then raise exception.Create('');
		ServerPort := StrToInt(Paramstr(2));
		if (ServerPort < 0) or (ServerPort > 65535) then raise exception.Create('');
	except
		WriteLn('Usage:   Client [IP | HostName] [Port]');
		WriteLn('Example: Client smtp.fibertel.com.ar 25');
		Halt(1);
	end;

    // Create an Adapter object representing out network adapter.
    Adapter := TPacketDriverAdapter.Create;

    // If the adapter works fine, we bind the TCP/IP protocol.
    if Adapter.Status <> asOk then begin
		WriteLn('Could not initialise network adapter. PKTPM is not loaded.');
		Halt(2);
	end;
    IPConf.DHCP      := False;
	IPConf.IPAddress := IP_StringToAddr(MY_IP);
	IPConf.NetMask   := IP_StringToAddr(MY_NETMASK);
    IPConf.Gateway   := IP_StringToAddr(MY_GATEWAY);
    IPHandle         := IP_Init(Adapter, IPConf);
	if IPHandle = 0 then begin
        WriteLn('Could not initialise IP stack. Wrong Parameters.');
        Adapter.Free;
		Halt(3);
	end;
	SetDNSServer(IP_StringToAddr(DNS_SERVER), Swap(WORD(53)));

	// Resolve Server IP
	ServerIP := IP_StringToAddr(ParamStr(1));
	if ServerIP = 0 then begin
		// Doesn't seem to be an IP. Let's try to resolve it with DNS.
		if not GetHostByName(ParamStr(1), ServerIP) then begin
			WriteLn('Could not resolve host name "' + ParamStr(1) + '".');
			IP_Deinit(IPHandle);
			Adapter.Free;
			Halt(4);
		end;
	end;

	// Create socket to connect to the Server
	HSocket := socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
	if HSocket <= 0 then begin
		WriteLn('Error creating socket: ', SocketError);
		IP_Deinit(IPHandle);
		Adapter.Free;
		Halt(4);
	end;

	// Set timeout to 10 seconds
	ConnTimeout := 10000;
	SetSocketOptions(HSocket, SOL_SOCKET, SOCKOPT_CONNTIMEOUT,
	  ConnTimeout, SizeOf(ConnTimeout));

	// Establish connection
	Write('Connecting to ' + IP_AddrToString(ServerIP) + ':' +
	  IntToStr(ServerPort) + '... ');
	SockAddr.family := AF_INET;
	SockAddr.addr   := ServerIP;
	SockAddr.port   := Swap(ServerPort);
	if not Connect(HSocket, SockAddr, SizeOf(SockAddr)) then begin
		// Could not connect
		WriteLn('Error ' + IntToStr(SocketError));
		CloseSocket(HSocket);
		IP_Deinit(IPHandle);
		Adapter.Free;
		Halt(5);
	end;
	WriteLn('Ok');

	// Loop until we are disconnected. When the user types something, we send it
	// to the server. If the server sends something, we show it on screen.
	While ioctlsocket(HSocket, FIONREAD, @ToRead) = 0 do begin
		// Any data to read from the server?
		if ToRead = 0 then begin
			// No data. At least, check to see if the server closed
			// the connection.
			if recv(HSocket, c, 0, 0) < 0 then Break;
		end else begin
			// Yes, there is!. Print it on screen.
			for i := 1 to ToRead do begin
				recv(HSocket, c, 1, 0);
				Write(c);
			end;
		end;
		// Something to send?
		if KeyPressed then begin
			c := ReadKey;
			if c = #27 then break;
			if c = #13 then begin
				c := #10;
				send(HSocket, c, 1, 0);
				c := #13;
			end;
			send(HSocket, c, 1, 0);
		end;
	end;
	CloseSocket(HSocket);

	// Close everything
	IP_Deinit(IPHandle);
	Adapter.Free;
end.
