unit Sockets;

Interface

Uses
     DOSSock;

Const
	{ Address Families }
    AF_MAX          = 12;
    PF_MAX          = AF_MAX;
    { Protocol Types }
	IPPROTO_IP     =   0;             { dummy for IP }
	IPPROTO_ICMP   =   1;             { control message protocol }
	IPPROTO_IGMP   =   2;             { group management protocol }
	IPPROTO_GGP    =   3;             { gateway^2 (deprecated) }
	IPPROTO_TCP    =   6;             { tcp }
	IPPROTO_PUP    =  12;             { pup }
	IPPROTO_UDP    =  17;             { user datagram protocol }
	IPPROTO_IDP    =  22;             { xns idp }
	IPPROTO_ND     =  77;             { UNOFFICIAL net disk proto }
	IPPROTO_RAW    =  255;            { raw IP packet }
    IPPROTO_MAX    =  256;
    { Address Constants }
    INADDR_ANY: DWORD = $00000000;
    INADDR_LOOPBACK: DWORD = $0100007F;
    INADDR_BROADCAST: DWORD = $FFFFFFFF;
    INADDR_NONE: DWORD = $FFFFFFFF;
    { Error codes }
	SOCKET_ERROR            = -1;
    INVALID_SOCKET          = SOCKET_ERROR;
    WSABASEERR              = 10000;
	WSAEINTR                = (WSABASEERR+4);
	WSAEBADF                = (WSABASEERR+9);
  	WSAEACCES               = (WSABASEERR+13);
  	WSAEFAULT               = (WSABASEERR+14);
  	WSAEINVAL               = (WSABASEERR+22);
  	WSAEMFILE               = (WSABASEERR+24);
  	WSAEWOULDBLOCK          = (WSABASEERR+35);
  	WSAEINPROGRESS          = (WSABASEERR+36);	
  	WSAEALREADY             = (WSABASEERR+37);
  	WSAENOTSOCK             = (WSABASEERR+38);
  	WSAEDESTADDRREQ         = (WSABASEERR+39);
  	WSAEMSGSIZE             = (WSABASEERR+40);
  	WSAEPROTOTYPE           = (WSABASEERR+41);
  	WSAENOPROTOOPT          = (WSABASEERR+42);
	WSAEPROTONOSUPPORT      = (WSABASEERR+43);
	WSAESOCKTNOSUPPORT      = (WSABASEERR+44);
	WSAEOPNOTSUPP           = (WSABASEERR+45);
	WSAEPFNOSUPPORT         = (WSABASEERR+46);
	WSAEAFNOSUPPORT         = (WSABASEERR+47);
	WSAEADDRINUSE           = (WSABASEERR+48);
	WSAEADDRNOTAVAIL        = (WSABASEERR+49);
	WSAENETDOWN             = (WSABASEERR+50);
  	WSAENETUNREACH          = (WSABASEERR+51);
	WSAENETRESET            = (WSABASEERR+52);
	WSAECONNABORTED         = (WSABASEERR+53);
	WSAECONNRESET           = (WSABASEERR+54);
	WSAENOBUFS              = (WSABASEERR+55);
	WSAEISCONN              = (WSABASEERR+56);
	WSAENOTCONN             = (WSABASEERR+57);
	WSAESHUTDOWN            = (WSABASEERR+58);
	WSAETOOMANYREFS         = (WSABASEERR+59);
	WSAETIMEDOUT            = (WSABASEERR+60);
  	WSAECONNREFUSED         = (WSABASEERR+61);
  	WSAELOOP                = (WSABASEERR+62);
  	WSAENAMETOOLONG         = (WSABASEERR+63);
  	WSAEHOSTDOWN            = (WSABASEERR+64);
  	WSAEHOSTUNREACH         = (WSABASEERR+65);
  	WSAENOTEMPTY            = (WSABASEERR+66);
  	WSAEPROCLIM             = (WSABASEERR+67);
  	WSAEUSERS               = (WSABASEERR+68);
  	WSAEDQUOT               = (WSABASEERR+69);
  	WSAESTALE               = (WSABASEERR+70);
  	WSAEREMOTE              = (WSABASEERR+71);
  	WSAEDISCON              = (WSABASEERR+101);
  	WSASYSNOTREADY          = (WSABASEERR+91);
  	WSAVERNOTSUPPORTED      = (WSABASEERR+92);
	WSANOTINITIALISED       = (WSABASEERR+93);
    { Various }
    IOCPARM_MASK = $7f;
    IOC_VOID     = $20000000;
    IOC_OUT      = $40000000;
    IOC_IN       = $80000000;
    IOC_INOUT    = (IOC_IN or IOC_OUT);
    FIONREAD     = IOC_OUT or { get # bytes to read }
      ((Longint(SizeOf(Longint)) and IOCPARM_MASK) shl 16) or
      (Longint(Byte('f')) shl 8) or 127;
    FIONBIO      = IOC_IN or { set/clear non-blocking i/o }
      ((Longint(SizeOf(Longint)) and IOCPARM_MASK) shl 16) or
      (Longint(Byte('f')) shl 8) or 126;
    FIOASYNC     = IOC_IN or { set/clear async i/o }
      ((Longint(SizeOf(Longint)) and IOCPARM_MASK) shl 16) or
      (Longint(Byte('f')) shl 8) or 125;
    MSG_PEEK        = $2;             { peek at incoming message }
    { Socket Options }
    SOL_SOCKET             = 1;
    SOCKOPT_CONNTIMEOUT    = 1;

{$I socketsh.inc}

{ Functions specific to DOS-Sockets implementation }
Function CloseSocket(Sock: LongInt): Longint;
Function ioctlsocket(Sock: LongInt; Option: LongInt; d: Pointer): LongInt;
//Function GetErrorMessage(ErrorCode: LongInt): String;

Implementation

{******************************************************************************
                          Basic Socket Functions
******************************************************************************}

Function socket(Domain, SocketType, Protocol: Longint): Longint;
begin
	Result := DOSSock.Socket(Domain, SocketType, ProtoCol);
	if Result < 0 then
		SocketError := DOSSock.GetLastError
	else
		SocketError:=0;
end;

Function CloseSocket(Sock: LongInt): Longint;
begin
    Result := DOSSock.CloseSocket(Sock);
	if Result < 0 then
		SocketError := DOSSock.GetLastError
	else
    	SocketError := 0;
end;

Function Send(Sock: Longint; Var Addr; AddrLen, Flags: Longint): Longint;
begin
	Result := DOSSock.Send(Sock, Addr, AddrLen, Flags);
    if Result < 0 then
		SocketError := DOSSock.GetLastError
	else
    	SocketError := 0;
end;

Function Recv(Sock: Longint; Var Addr; AddrLen, Flags: Longint): Longint;
begin
	Result := DOSSock.Recv(Sock, Addr, AddrLen, Flags);
	if Result < 0 then
    	SocketError := DOSSock.GetLastError
	else
    	SocketError := 0;
end;

Function Bind(Sock: Longint; Var Addr; AddrLen: Longint): Boolean;
begin
    if DOSSock.Bind(Sock, Addr, AddrLen) then begin
       SocketError := 0;
       Bind := true;
    end else begin
       SocketError := DOSSock.GetLastError;
       Bind := false;
    end;
end;

Function Listen(Sock, MaxConnect: Longint): Boolean;
begin
    if DOSSock.Listen(Sock,MaxConnect) then begin
       SocketError := 0;
       Listen := true;
    end else begin
       SocketError := DOSSock.GetLastError;
       Listen := false;
    end
end;

Function Accept(Sock: Longint; Var Addr; Var Addrlen: Longint): Longint;
begin
    Result := DOSSock.Accept(Sock, Addr, AddrLen);
    if Result < 0 then
		SocketError := DOSSock.GetLastError
	else
    	SocketError := 0;
end;

Function Connect(Sock: Longint; Var Addr; Addrlen: Longint): Boolean;
begin
    Result := DOSSock.Connect(Sock, Addr, AddrLen);
    if not Result then
	    SocketError := DOSSock.GetLastError
  	else
    	SocketError := 0;
end;

Function Shutdown(Sock: Longint; How: Longint): Longint;
begin
    Result := DOSSock.ShutDown(Sock, How);
    if Result < 0 then
    	SocketError := DOSSock.GetLastError
  	else
    	SocketError:=0;
end;

Function GetSocketName(Sock: Longint; Var Addr; Var Addrlen: Longint): Longint;
begin
    Result := DOSSock.GetSocketName(Sock, Addr, AddrLen);
    if Result < 0 then
    	SocketError := DOSSock.GetLastError
  	else
    	SocketError := 0;
end;

Function GetPeerName(Sock: Longint; Var Addr; Var Addrlen: Longint): Longint;
begin
    Result := DOSSock.GetPeerName(Sock, Addr, AddrLen);
    if Result < 0 then
		SocketError := DOSSock.GetLastError
	else
    	SocketError := 0;
end;

Function SetSocketOptions(Sock, Level, OptName: Longint; Var OptVal; OptLen: Longint): Longint;
begin
    Result := DOSSock.SetSocketOptions(Sock, Level, OptName, OptVal, OptLen);
    if Result < 0 then
    	SocketError := DOSSock.GetLastError
	else
    	SocketError := 0;
end;

Function GetSocketOptions(Sock, Level, OptName: Longint; Var OptVal; Var OptLen: Longint): Longint;
begin
    Result := DOSSock.GetSocketOptions(Sock, Level, OptName, OptVal, OptLen);
    if Result < 0 then
	    SocketError := DOSSock.GetLastError
	else
    	SocketError := 0;
end;

Function SocketPair(Domain, SocketType, Protocol: Longint; var Pair: TSockArray): Longint;
begin
    // SocketPair:=SocketCall(Socket_Sys_SocketPair,Domain,SocketType,Protocol,longint(@Pair),0,0);
    Result := SOCKET_ERROR;
end;

Function ioctlsocket(Sock: LongInt; Option: LongInt; d: Pointer): LongInt;
begin
    Result := DOSSock.ioctlsocket(Sock, Option, d);
    if Result < 0 then
    	SocketError := DOSSock.GetLastError
	else
        SocketError := 0;
end;

{ mimic the linux fdWrite/fdRead calls for the file/text socket wrapper }
function fdWrite(handle: longint; var bufptr; size: dword): LongInt;
begin
    Result := DOSSock.send(handle, bufptr, size, 0);
    if Result = SOCKET_ERROR then begin
	    SocketError := DOSSock.GetLastError;
	    fdWrite := 0;
    end else
	    SocketError := 0;
end;

function fdRead(handle : longint;var bufptr;size : dword) : LongInt;
var
	d: dword;
begin
    if ioctlsocket(handle, FIONREAD, @d) = SOCKET_ERROR then begin
		SocketError := DOSSock.GetLastError;
        fdRead := 0;
    	exit;
    end;
    if d > 0 then begin
        if size > d then size := d;
        Result := DOSSock.recv(handle, bufptr, size, 0);
        if Result = SOCKET_ERROR then begin
        	SocketError := DOSSock.GetLastError;
        	fdRead := 0;
        end else
           SocketError:=0;
    end else
       SocketError:=0;
end;
(*
Function TCP_GetErrorMessage(ErrorCode: Integer): String;
begin
	Case ErrorCode of
		WSAEINTR                = (WSABASEERR+4);
		WSAEBADF                = (WSABASEERR+9);
  		WSAEACCES               = (WSABASEERR+13);
  		WSAEFAULT               = (WSABASEERR+14);
  		WSAEINVAL               = (WSABASEERR+22);
  		WSAEMFILE               = (WSABASEERR+24);
  		WSAEWOULDBLOCK          = (WSABASEERR+35);
  		WSAEINPROGRESS          = (WSABASEERR+36);	
  		WSAEALREADY             = (WSABASEERR+37);
  		WSAENOTSOCK             = (WSABASEERR+38);
  	WSAEDESTADDRREQ         = (WSABASEERR+39);
  	WSAEMSGSIZE             = (WSABASEERR+40);
  	WSAEPROTOTYPE           = (WSABASEERR+41);
  	WSAENOPROTOOPT          = (WSABASEERR+42);
	WSAEPROTONOSUPPORT      = (WSABASEERR+43);
	WSAESOCKTNOSUPPORT      = (WSABASEERR+44);
	WSAEOPNOTSUPP           = (WSABASEERR+45);
	WSAEPFNOSUPPORT         = (WSABASEERR+46);
	WSAEAFNOSUPPORT         = (WSABASEERR+47);
	WSAEADDRINUSE           = (WSABASEERR+48);
	WSAEADDRNOTAVAIL        = (WSABASEERR+49);
	WSAENETDOWN             = (WSABASEERR+50);
  	WSAENETUNREACH          = (WSABASEERR+51);
	WSAENETRESET            = (WSABASEERR+52);
	WSAECONNABORTED         = (WSABASEERR+53);
	WSAECONNRESET           = (WSABASEERR+54);
	WSAENOBUFS              = (WSABASEERR+55);
	WSAEISCONN              = (WSABASEERR+56);
	WSAENOTCONN             = (WSABASEERR+57);
	WSAESHUTDOWN            = (WSABASEERR+58);
	WSAETOOMANYREFS         = (WSABASEERR+59);
	WSAETIMEDOUT            = (WSABASEERR+60);
  	WSAECONNREFUSED         = (WSABASEERR+61);
  	WSAELOOP                = (WSABASEERR+62);
  	WSAENAMETOOLONG         = (WSABASEERR+63);
  	WSAEHOSTDOWN            = (WSABASEERR+64);
  	WSAEHOSTUNREACH         = (WSABASEERR+65);
  	WSAENOTEMPTY            = (WSABASEERR+66);
  	WSAEPROCLIM             = (WSABASEERR+67);
  	WSAEUSERS               = (WSABASEERR+68);
  	WSAEDQUOT               = (WSABASEERR+69);
  	WSAESTALE               = (WSABASEERR+70);
  	WSAEREMOTE              = (WSABASEERR+71);
  	WSAEDISCON              = (WSABASEERR+101);
  	WSASYSNOTREADY          = (WSABASEERR+91);
  	WSAVERNOTSUPPORTED      = (WSABASEERR+92);
	WSANOTINITIALISED       = (WSABASEERR+93);

        SCK_INVALID_SOCKET:
        	TCP_GetErrorMessage := 'Socket invlido';
	    SCK_NO_MORE_SOCKETS:
        	TCP_GetErrorMessage := 'No hay ms sockets disponibles';
    	SCK_NOT_CLOSED:
			TCP_GetErrorMessage := 'No se puede ejecutar la operacin porque el socket no est cerrado';
	    SCK_NOT_BOUND:
        	TCP_GetErrorMessage := 'El socket no est unido a un IP y un Port';
	    SCK_NOT_LISTENING:
        	TCP_GetErrorMessage := 'El socket no est en modo de escucha';
	    SCK_NOTHING_TO_ACCEPT:
        	TCP_GetErrorMessage := 'No hay conexiones para aceptar';
	    SCK_TIMEOUT:
        	TCP_GetErrorMessage := 'Timeout en la operacin';
	    SCK_NOT_CONNECTED:
        	TCP_GetErrorMessage := 'No se puede ejecutar la operacin porque el socket no est conectado';
	    SCK_CONNECTION_REFUSED:
        	TCP_GetErrorMessage := 'La conexin ha sido rechazada por el equipo remoto';
        else
        	TCP_GetErrorMessage := 'Error desconocido';
    end;
end;

*)

end.
