Unit DNS;

interface

Uses
    SysUtils, Sockets, Network, IP, UDP, Util;

Procedure SetDNSServer(ServerIP: TIPAddress; Port: WORD);
Function  GetHostByName(HostName: AnsiString; var IPAddress: TIPAddress): Boolean;

implementation

Var
	DNS_PORT: WORD;
	DNS_IP: TIPAddress;

Type
	PWORD = ^WORD;

	PDNSHeader = ^TDNSHeader;
	TDNSHeader = packed record
    	ID: WORD;
        Param0: BYTE;
        Param1: BYTE;
        RequestCount: WORD;
        AnswerCount: WORD;
        AuthCount: WORD;
        AdditCount: WORD;
    end;

    TDNSServers = record
    	Count: Integer;
        Server: Array[1..20] of String;
    end;

    TDNSCacheEntry = record
    	HostName: String;
        IPAddress: TIPAddress;
        Expires: Comp;
    end;

Var
    Cache: Array[1..20] of TDNSCacheEntry;

Procedure SetDNSServer(ServerIP: TIPAddress; Port: WORD);
begin
	DNS_IP   := ServerIP;
	DNS_PORT := Port;
end;

{**************************************************************************}
{                                                                          }
{                       Manejo del Cache de DNS                            }
{                                                                          }
{**************************************************************************}
function Cache_GetHostByName(HostName: AnsiString; Var IPAddress: TIPAddress): Boolean;
Var
	i: Integer;
begin
    { Pasamos el HostName a maysculas }
    HostName := UpperCase(HostName);
	{ Buscamos por todo el Cache, teniento en cuenta slo las entradas }
	{ que an no expiraron.                                            }
	for i := Low(Cache) to High(Cache) do begin
        if (Cache[i].HostName = HostName)
          and (Cache[i].Expires > GetMSCounter) then begin
          	{ Lo encontramos ! }
            IPAddress := Cache[i].IPAddress;
          	Cache_GetHostByName := True;
          	Exit;
        end;
    end;
    { No lo encontramos }
    Cache_GetHostByName := False;
end;

procedure Cache_AddHostByName(HostName: AnsiString; IPAddress: TIPAddress);
Var
	i, Entry: Integer;
	TmpIP: TIPAddress;
begin
	{ Si ya est en la lista, no hacemos nada }
    if Cache_GetHostByName(HostName, TmpIP) then Exit;
	{ Pasamos el HostName a maysculas }
    HostName := UpperCase(HostName);
    { Quitamos las entradas caducas del Cache }
	for i := Low(Cache) to High(Cache) do begin
    	if (Cache[i].HostName <> '') and (Cache[i].Expires < GetMSCounter) then begin
            Cache[i].HostName := '';
        end;
    end;
    { Buscamos una entrada libre en el Cache }
    Entry := Random(High(Cache)) + 1;
	for i := Low(Cache) to High(Cache) do begin
    	if (Cache[i].HostName = '') then begin
        	Entry := i;
            Break;
        end;
    end;
	{ Copiamos los datos a la entrada elegida }
    Cache[Entry].IPAddress := IPAddress;
    Cache[Entry].HostName  := HostName;
    Cache[Entry].Expires   := GetMSCounter + 40000;		{ 40 segundos }
end;


{**************************************************************************}
{                                                                          }
{                Procesamiento de Paquetes del Server DNS                  }
{                                                                          }
{**************************************************************************}
Function GetResourceName(Org: PChar; var P: PChar): AnsiString;
Var
	b: BYTE;
	q: PChar;
	s: String;
    i: Integer;
    Comp: Boolean;
begin
   	s := '';
    Comp := False;
	While P[0] <> #0 do begin
    	B := BYTE(P[0]);
        if s <> '' then s := s + '.';
        if (B and 192) = 192 then begin
        	Comp := True;
        	Q := Org + (B and 63) * $100 + BYTE(P[1]);
            s := s + GetResourceName(Org, Q);
            Inc(P, 2);
            Break;
        end else begin
            for i := 1 to B do s := s + P[i];
            Inc(P, B + 1);
        end;
	end;
    if not Comp then Inc(P);
	GetResourceName := s;
end;

function MakeResourceName(s: AnsiString): AnsiString;
Var
	i, n, q: Integer;
begin
    s := '.' + s;
	q := 1;
	n := 0;
    for i := 1 to Length(s) do begin
		if (s[i] = '.') then begin
        	s[q] := Chr(n);
            q := i;
            n := 0;
        end else begin
            Inc(n);
        end;
    end;
    s[q] := Chr(n);
    MakeResourceName := s;
end;

procedure ProcessReq(Org: PChar; Var P: PChar);
begin
	GetResourceName(Org, P);
    Inc(P, 4);
end;

procedure ProcessAns(
  Org: PChar;
  Var DNSServers: TDNSServers;
  Var CanonicalName: AnsiString;
  Var P: PChar);
Var
//	TTL: DWORD;
    DNS2: PChar;
    ResourceName: String;
    ResType, {ResClass, }DataLen: WORD;
begin
(*	WriteLn('Answer');	*)
    { Procesamos el Nombre del Recurso }
    ResourceName := GetResourceName(Org, P);
    { Procesamos el Tipo, Clase, TTL y DataLen del Recurso }
    ResType  := Swap(PWORD(P)^);
    Inc(P, 2);
//    ResClass := Swap(PWORD(P)^);
	Inc(P, 2);
//    TTL      := SwapDWORD(PDWORD(P)^);
	Inc(P, 4);
    DataLen  := Swap(PWORD(P)^);
    Inc(P, 2);
    { Obtenemos los Datos del Recurso }
{	WriteLn('Resource Name    : ', ResourceName);
	WriteLn('Respuesta de tipo: ', ResType);
	WriteLn('Clase            : ', ResClass);
    WriteLn('DataLen          : ', DataLen);
}
    case ResType of
    	1:  { Direccin IP de un equipo }
        	begin
//          	WriteLn('***** IP *****');
				Cache_AddHostByName(ResourceName, PIPAddress(P)^);
            end;

        2:	{ Servidor de Dominio }
        	begin
//                WriteLn('***** DNS *****');
                if DNSServers.Count < High(DNSServers.Server) then begin
	                DNS2 := P;
    	            Inc(DNSServers.Count);
                    DNSServers.Server[DNSServers.Count] :=
                      GetResourceName(Org, DNS2);
                end;
            end;
        5:	{ Forma cannica de un Alias }
        	begin
//         		WriteLn('***** Forma cannica *****');
                DNS2 := P;
                CanonicalName := GetResourceName(Org, DNS2);
//                WriteLn('Nombre Cannico: ', CanonicalName);
            end;
    end;
    P := @P[DataLen];
end;

procedure ProcessDNSRequest(
  P: PDNSHeader;
  Var DNSServers: TDNSServers;
  Var CanonicalName: AnsiString);
Var
    PC: PChar;
	i: Integer;
begin
(*    for i := 0 to 150 do begin
    	c := PChar(P)[i];
        case c of
        	#0..#32: Write('<', ord(c), '>');
            else Write(c);
        end;
	end;
    WriteLn;	*)
	{ Procesamos el Paquete de respuesta DNS }
    PC := PChar(P) + SizeOf(TDNSHeader);
{	WriteLn('Param     : ', P^.Param0, ',', P^.Param1);
	WriteLn('Preguntas : ', Swap(P^.RequestCount));
	WriteLn('Respuestas: ', Swap(P^.AnswerCount));
	WriteLn('Auth      : ', Swap(P^.AuthCount));
	WriteLn('Additional: ', Swap(P^.AdditCount));
}
    { Procesamos las Preguntas }
    for i := 1 to Swap(P^.RequestCount) do ProcessReq(PChar(P), PC);
    { Procesamos las Respuestas }
    for i := 1 to Swap(P^.AnswerCount) do ProcessAns(PChar(P), DNSServers, CanonicalName, PC);
    { Procesamos las autorizaciones }
    for i := 1 to Swap(P^.AuthCount) do ProcessAns(PChar(P), DNSServers, CanonicalName, PC);
   	{ Procesamos los Adicionales }
    for i := 1 to Swap(P^.AdditCount) do ProcessAns(PChar(P), DNSServers, CanonicalName, PC);
end;

Function QueryDNSServers(NS_IP: TIPAddress; HostName: AnsiString; var DNSServers: TDNSServers): Boolean;
{
  Esta Funcin devuelve una lista de Servidores DNS capaces de resolver
  un nombre de Host dado. Si no hay ningn server que pueda resolver este
  Host, entonces devuelve False.
}
Var
    P: PChar;
	i: Integer;
	Timeout: Comp;
	RcvLen: DWORD;
	HSocket: Pointer;
	ServerOk: Boolean;
	ResourceName: String;
	SockAddr: TInetSockAddr;
    ReqType, ReqClass: WORD;
    CanonicalName: AnsiString;
    DNSRequest, DNSReply: PDNSHeader;
begin
    { Pedimos memoria para el Pedido }
    if not GetMemCheck(Pointer(DNSRequest),
	  SizeOf(TDNSHeader) + 1 + Length(HostName) + 1 + 2 + 2) then Exit;
    { Armamos el Header }
	DNSRequest^.ID := Random($FFFF) + 1;
   	DNSRequest^.Param0 := 1;
    DNSRequest^.Param1 := 0;
   	DNSRequest^.RequestCount := Swap(Word(1));
    DNSRequest^.AnswerCount := 0;
   	DNSRequest^.AuthCount := 0;
    DNSRequest^.AdditCount := 0;
   	ReqType := Swap(Word(2));			{ Pedimos Servidores DNS }
    ReqClass := Swap(Word(1));
    { Armamos los Datos }
    P := PChar(DNSRequest) + SizeOf(TDNSHeader);
    ResourceName := MakeResourceName(HostName);
    Move(ResourceName[1], P[0], Length(ResourceName)); Inc(P, Length(ResourceName));
   	P[0] := #0; Inc(P, 1);
    Move(ReqType, P[0], SizeOf(ReqType)); Inc(P, SizeOf(ReqType));
   	Move(ReqClass, P[0], SizeOf(ReqClass));
    { Abrimos un Socket para comunicarnos con el Servidor DNS }
    HSocket := UDP_Socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
    { Hacemos Bind }
    SockAddr.family := AF_INET;
    SockAddr.addr := INADDR_ANY;
    SockAddr.Port := 0;
    UDP_Bind(HSocket, SockAddr, SizeOf(SockAddr));
    { Conectamos al equipo remoto }
    SockAddr.family := AF_INET;
    SockAddr.addr := NS_IP;
    SockAddr.Port := DNS_PORT;
    UDP_Connect(HSocket, SockAddr, SizeOf(SockAddr));
    { Esperamos la respuesta }
	ServerOk := False;
	for i := 1 to 2 do begin		{ 2 intentos de 5 seg. con este server }
    	{ Enviamos el Pedido }
        if UDP_Send(HSocket, DNSRequest^, SizeOf(TDNSHeader) + 1 +
          Length(HostName) + 1 + 2 + 2, 0) = SOCKET_ERROR then Break;
        { Esperamos la respuesta }
        Timeout := GetMSCounter + 5000;     // 5 segundos
        While (GetMSCounter < Timeout) and not ServerOk do begin
            UDP_ioctlsocket(HSocket, FIONREAD, @RcvLen);
            if RcvLen > 0 then begin
                { Lleg la respuesta. Armamos un Buffer, y la copiamos }
	            GetMem(DNSReply, RcvLen);
                UDP_Recv(HSocket, DNSReply^, RcvLen, 0);
    		    ProcessDNSRequest(DNSReply, DNSServers, CanonicalName);
	            FreeMem(DNSReply, RcvLen);
	            ServerOk := True;
            end;
        end;
        if ServerOk then Break;
	end;
    { Liberamos la memoria del Pedido }
	FreeMem(DNSRequest, SizeOf(TDNSHeader) + 1 + Length(HostName) + 1 + 2 + 2);
    { Cerramos el Socket }
	UDP_Close(HSocket);
    QueryDNSServers := (DNSServers.Count > 0);
end;


Function QueryHostAddress(NS_IP: TIPAddress;
  HostName: AnsiString; Var CanonicalName: AnsiString): Boolean;
Var
    P: PChar;
	i: Integer;
	Timeout: Comp;
	RcvLen: DWORD;
	HSocket: Pointer;
	ServerOk: Boolean;
	ResourceName: String;
	SockAddr: TInetSockAddr;
    DNSServers: TDNSServers;
    HostAddress: TIPAddress;
    ReqType, ReqClass: WORD;
    DNSRequest, DNSReply: PDNSHeader;
begin
    QueryHostAddress := False;
    { Pedimos memoria para el Pedido }
	if not GetMemCheck(Pointer(DNSRequest),
	  SizeOf(TDNSHeader) + 1 + Length(HostName) + 1 + 2 + 2) then Exit;
    { Armamos el Header }
	DNSRequest^.ID := Random($FFFF) + 1;
   	DNSRequest^.Param0 := 1;
    DNSRequest^.Param1 := 0;
   	DNSRequest^.RequestCount := Swap(Word(1));
    DNSRequest^.AnswerCount := 0;
   	DNSRequest^.AuthCount := 0;
    DNSRequest^.AdditCount := 0;
   	ReqType := Swap(Word(1));			{ Pedido de Address }
    ReqClass := Swap(Word(1));
    { Armamos los Datos }
    P := PChar(DNSRequest) + SizeOf(TDNSHeader);
    ResourceName := MakeResourceName(HostName);
    Move(ResourceName[1], P[0], Length(ResourceName)); Inc(P, Length(ResourceName));
   	P[0] := #0; Inc(P, 1);
    Move(ReqType, P[0], SizeOf(ReqType)); Inc(P, SizeOf(ReqType));
   	Move(ReqClass, P[0], SizeOf(ReqClass));
    { Abrimos un Socket para comunicarnos con el Servidor DNS }
    HSocket := UDP_Socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
    { Hacemos Bind }
    SockAddr.family := AF_INET;
    SockAddr.addr := INADDR_ANY;
    SockAddr.Port := 0;
    UDP_Bind(HSocket, SockAddr, SizeOf(SockAddr));
    { Conectamos al equipo remoto }
    SockAddr.family := AF_INET;
    SockAddr.addr := NS_IP;
    SockAddr.Port := DNS_PORT;
    UDP_Connect(HSocket, SockAddr, SizeOf(SockAddr));
    { Esperamos la respuesta }
	ServerOk := False;
	for i := 1 to 2 do begin		{ 2 intentos de 5 seg. con este server }
    	{ Enviamos el Pedido }
        if UDP_Send(HSocket, DNSRequest^, SizeOf(TDNSHeader) + 1 +
          Length(HostName) + 1 + 2 + 2, 0) = SOCKET_ERROR then Break;
        { Esperamos la respuesta }
        Timeout := GetMSCounter + 5000;
        While (GetMSCounter < Timeout) and not ServerOk do begin
            UDP_ioctlsocket(HSocket, FIONREAD, @RcvLen);
            if RcvLen > 0 then begin
                { Lleg la respuesta. Armamos un Buffer, y la copiamos }
    	        if MaxAvail > RcvLen + 1024 then begin
	               	GetMem(DNSReply, RcvLen);
                    UDP_Recv(HSocket, DNSReply^, RcvLen, 0);
                    CanonicalName := '';
                    ProcessDNSRequest(DNSReply, DNSServers, CanonicalName);
                    { Recibimos una respuesta }
                    QueryHostAddress := (CanonicalName <> '')
                      or Cache_GetHostByName(HostName, HostAddress);
                    FreeMem(DNSReply, RcvLen);
                end;
                ServerOk := True;
    	    end;
	    end;
        if ServerOk then Break;
	end;
    { Liberamos la memoria del Pedido }
    FreeMem(DNSRequest, SizeOf(TDNSHeader) + 1 + Length(HostName) + 1 + 2 + 2);
    { Cerramos el Socket }
    UDP_Close(HSocket);
end;

Function GetHostByName(HostName: AnsiString; var IPAddress: TIPAddress): Boolean;
Var
    i: Integer;
    DNSServers: TDNSServers;
    DNSServerIP: TIPAddress;
    CanonicalName, SearchName: AnsiString;
begin
	GetHostByName := False;
	{ Si la direccin IP del Server ya est en Cache, la devolvemos }
    { directamente.                                                 }
    if Cache_GetHostByName(HostName, IPAddress) then begin
        GetHostByName := True;
    	Exit;
    end;
    { Armamos una lista de Servidores DNS para resolver nuestro pedido }
    DNSServers.Count := 0;
    QueryDNSServers(DNS_IP, HostName, DNSServers);
    { Probamos en cada servidor DNS, buscando si puede resolver el }
    { nombre del Host.                                             }
    i := 0;
    CanonicalName := '';
    SearchName := HostName;
    While not Cache_GetHostByName(SearchName, IPAddress)
      and (i <= DNSServers.Count) do begin
        if i = 0 then
            DNSServerIP := DNS_IP
        else begin
        	if not Cache_GetHostByName(DNSServers.Server[i], DNSServerIP) then begin
                DNSServerIP := 0;
            end;
        end;
        if DNSServerIP <> 0 then begin
            { Le pedimos el IP del equipo a este Server }
//			if i > 0 then WriteLn('Probando desde ', DNSServers.Server[i]);
            if QueryHostAddress(DNSServerIP, HostName, CanonicalName) then begin
                if CanonicalName <> '' then begin
//					WriteLn('Nombre Cannico: ', CanonicalName);
                    SearchName := CanonicalName;
                    i := 0;
                end;
            end;
        end;
        Inc(i);
    end;
    { Listo }
    if Cache_GetHostByName(SearchName, IPAddress) then begin
        Cache_AddHostByName(HostName, IPAddress);
        GetHostByName := True;
    end else begin
        GetHostByName := False;
    end;
end;


Var
	i: Integer;
initialization
	DNS_PORT := 0;
	DNS_IP := 0;
	for i := Low(Cache) to High(Cache) do Cache[i].HostName := '';
end.
