Unit IP;

interface

Uses
    Util, Network, ARP, Objects;

Type
    { 4 Byte IP address}
    PIPAddress = ^TIPAddress;
    TIPAddress = DWORD;

    { Rutina para procesar paquetes IP }
    TIP_Routine = procedure(SourceIP, TargetIP: TIPAddress;
      Var Data; DataLen: Word);

    TIPAdapterConfig = record
    	DHCP: Boolean;
        IPAddress: TIPAddress;
        NetMask: TIPAddress;
        Gateway: TIPAddress;
	end;

Function   IP_Init(Adapter: TNetworkAdapter; Config: TIPAdapterConfig): DWORD;
Function   IP_Deinit(IPHandle: DWORD): Boolean;
Function   IP_AddProcessor(Protocol: BYTE; Routine: TIP_Routine): Boolean;
Function   IP_RemoveProcessor(Protocol: BYTE): Boolean;
Function   IP_AddrToString(IP: TIPAddress): String;
Function   IP_StringToAddr(StrAddr: String): TIPAddress;
Function   IP_Checksum(Data: Pointer; DataCnt: WORD; PIP: pointer; PIPCnt: WORD): WORD;
Function   IP_Send(TargetIP: TIPAddress; Protocol: BYTE; Var Data; DataLen: WORD): LongInt;
Function   IP_AddRoute(Destination, Mask, Gateway, LocalIP: TIPAddress; Cost: Integer): Boolean;
Function   IP_GetRouteForTarget(TargetIP: TIPAddress; var LocalIP, Gateway: TIPAddress): Boolean;
Procedure  IP_PrintRoute;
Function   IP_GetDHCPOption(Option: BYTE; Var Buffer; Var BuffLen: LongInt): Boolean;
Function   IP_GetIPByHandle(IPHandle: DWORD): TIPAddress;

implementation

Uses
	Sockets, ICMP, UDP, TCP, DHCP;

Const
	TYPELEN_IP = Swap(Word($0800));

Type
	PIPHeader = ^TIPHeader;
	TIPHeader = packed record
        VersHLen: BYTE;
        ServiceType: BYTE;
        TotalLength: WORD;
        ID: WORD;
        FlagsFrag1: BYTE;
        Frag2: BYTE;
        LifeTime: BYTE;
        Protocol: BYTE;
        Checksum: WORD;
        SourceIP: TIPAddress;
        TargetIP: TIPAddress;
    end;

    PAddressEntry = ^TAddressEntry;
    TAddressEntry = record
    	// Basic parameters
        IPAddress: TIPAddress;			// IP Address
        Adapter: TNetworkAdapter;      // Adapter
		// DHCP-related parameters
        DHCPInUse: Boolean;				// Are we using DHCP for this address ?
        DHCPServer: TIPAddress;         // DHCP Server used to get it
        DHCPT1: Comp;                 	// Renewal Time
        DHCPT2: Comp;					// Rebinding Time
        DHCPT3: Comp; 					// End-of-lease Time
    end;

	PRouteEntry = ^TRouteEntry;
    TRouteEntry = record
        Destination: TIPAddress;
        Mask: TIPAddress;
        Gateway: TIPAddress;
        LocalIP: TIPAddress;
        Cost: Integer;
    end;

    TCacheEntry = record
        Expires: Comp;
    	IPAddress: TIPAddress;
        EthernetAddress: TEthernetAddress;
    end;

    TIPProcessors = record
    	Protocol: BYTE;
        Routine: TIP_Routine;
    end;

    TFragment = record
    	Offset: WORD;
    	Data: Pointer;
        DataLen: WORD;
    end;

    TDatagram = record
    	SourceIP: TIPAddress;
        TargetIP: TIPAddress;
        Id: WORD;
        Protocol: BYTE;
        FragCount: BYTE;
		TotalLength: WORD;
        ReceivedLength: WORD;
        Timeout: Comp;
		Fragments: Array[1..40] of TFragment;
    end;

Var
    CheckingDHCP: Boolean;
    RouteList: PCollection;
    AddressList: PCollection;
    Datagrams: Array[1..10] of TDatagram;
    CacheList: Array[1..100] of TCacheEntry;
    Processors: Array[1..10] of TIPProcessors;

procedure IP_ProcessPacket(Adapter: TNetworkAdapter; Packet: PEthernetPacket); forward;

{**************************************************************************}
{                                                                          }
{                   Manejo de Adaptadores de Red para IP                   }
{                                                                          }
{**************************************************************************}
Function IP_GetIPByHandle(IPHandle: DWORD): TIPAddress;
Var
    AE: PAddressEntry;
begin
    Result := 0;
    AE := PAddressEntry(IPHandle);
    if AddressList^.IndexOf(AE) = -1 then Exit;
    Result := AE^.IPAddress;
end;


Function GetAdapterFromAddress(IPAddress: TIPAddress): TNetworkAdapter;
Var
    i: Integer;
    P: PAddressEntry;
begin
    for i := 0 to AddressList.Count - 1 do begin
        P := AddressList^.At(i);
        if (P^.IPAddress = IPAddress) then begin
        	GetAdapterFromAddress := P^.Adapter;
        	Exit;
        end;
    end;
    GetAdapterFromAddress := nil;
end;

procedure IP_CheckDHCPEntry(var Entry: TAddressEntry);
Var
    DW: DWORD;
	l: LongInt;
    i: Integer;
    Timeout: Comp;
    RE: PRouteEntry;
    NetMask, Gateway: TIPAddress;
begin
    // If the entry is still valid, there's nothing to do
    if (Entry.DHCPT1 > GetMSCounter) and (Entry.DHCPT2 > GetMSCounter)
      and (Entry.DHCPT3 > GetMSCounter) then Exit;
    // Es hora de renovar nuestro IP
    if Entry.IPAddress = 0 then begin
        // Este adaptador perdi su direccin IP.
        // Hay que nogociarla de cero.
        if DHCP_ConfigAdapter(Entry.Adapter, Entry.IPAddress,
          NetMask, Entry.DHCPServer, DW) then begin
            Timeout := DW;
            Timeout := Timeout * 1000;
            Entry.DHCPT1 := GetMSCounter + (Timeout * 0.500);
            Entry.DHCPT2 := GetMSCounter + (Timeout * 0.875);
            Entry.DHCPT3 := GetMSCounter + Timeout;
         (* WriteLn(Entry.DHCPT1);
            WriteLn(Entry.DHCPT2);
            WriteLn(Entry.DHCPT3); *)
          // Agregamos la ruta de loopback para el adaptador
            IP_AddRoute(Entry.IPAddress, $FFFFFFFF, INADDR_LOOPBACK, INADDR_LOOPBACK, 0);
            // Agregamos la ruta para la red local del adaptador
            IP_AddRoute(Entry.IPAddress and NetMask, NetMask,
              Entry.IPAddress, Entry.IPAddress, 0);
            // Intentamos obtener un Gateway
            l := SizeOf(Gateway);
            if IP_GetDHCPOption(3, Gateway, l) then begin
                IP_AddRoute(0, 0, Gateway, Entry.IPAddress, 9999);
            end;
{            WriteLn('Se volvi a conseguir un IP');}
        end else begin
            Entry.IPAddress := 0;
            Entry.DHCPT1    := 99999999999999;
            Entry.DHCPT2    := 99999999999999;
            Entry.DHCPT3    := GetMSCounter + 60000;    // Reintentar en 1 minuto
//          WriteLn('NO se volvi a conseguir un IP');
        end;
    end else begin
    	// Intentamos renovar la direccin IP
        if DHCP_RenewIP(Entry.Adapter, Entry.IPAddress,
          Entry.DHCPServer, DW) then begin
            Timeout := DW;
    	    Timeout := Timeout * 1000;
        	Entry.DHCPT1 := GetMSCounter + (Timeout * 0.500);
	        Entry.DHCPT2 := GetMSCounter + (Timeout * 0.875);
            Entry.DHCPT3 := GetMSCounter + Timeout;
//          WriteLn('Se renov el IP');
            Exit;
        end;
        // No se pudo renovar la direccin IP. En qu punto estbamos ?
        if Entry.DHCPT1 < GetMSCounter then begin
	        // Fallamos al renovar en T1
            Entry.DHCPT1 := 99999999999999;
//          WriteLn('No se renov el IP en T1');
        end else if Entry.DHCPT2 < GetMSCounter then begin
	        // Fallamos al renovar en T2
            Entry.DHCPT2 := 99999999999999;
//          WriteLn('No se renov el IP en T2');
        end else if Entry.DHCPT3 < GetMSCounter then begin
            // Fallamos al renovar en T3. El IP ya no es vlido.
            // Eliminamos los items de la tabla de ruteo que corresponden a
            // este IP
            i := 0;
            While i < RouteList^.Count do begin
                RE := RouteList^.At(i);
                if (RE^.Destination = Entry.IPAddress) or
                  (RE^.LocalIP = Entry.IPAddress) then begin
                  	// Hay que eliminar este item
                    RouteList.AtDelete(i);
                    Dispose(RE);
                end else begin
                	Inc(i);
                end;
            end;
            // Ponemos esta direccin en cero.
            Entry.DHCPT1    := 99999999999999;
            Entry.DHCPT2    := 99999999999999;
            Entry.DHCPT3    := GetMSCounter + 60000;    // Reintentar en 1 minuto
            Entry.IPAddress := 0;
//          WriteLn('Se perdi el IP');
    	end;
    end;
end;

procedure IP_CheckDHCP;
Var
	i: Integer;
	AE: PAddressEntry;
begin
	CheckingDHCP := True;
    for i := 0 to AddressList^.Count - 1 do begin
        AE := AddressList^.At(i);
        if AE^.DHCPInUse then IP_CheckDHCPEntry(AE^);
    end;
    CheckingDHCP := False;
end;


{**************************************************************************}
{                                                                          }
{                          Manejo de ARP para IP                           }
{                                                                          }
{**************************************************************************}
function GetEthernetAddress(IPAddress: TIPAddress;
  var EthernetAddress: TEthernetAddress): Boolean;
(*
  Esta rutina busca en el Cache de direcciones de hardware la direccin de
  un IP dado. Si la direccin no est en el Cache, devuelve False.
*)
Var
    i: integer;
begin
	GetEthernetAddress := True;
    for i := Low(CacheList) to High(CacheList) do begin
        if (CacheList[i].Expires + 0 > GetMSCounter) and
          (CacheList[i].IPAddress = IPAddress) then begin
            EthernetAddress := CacheList[i].EthernetAddress;
            Exit;
		end;
    end;
    GetEthernetAddress := False;
end;

procedure SetEthernetAddress(IPAddress: TIPAddress;
  EthernetAddress: TEthernetAddress);
(*
  Esta rutina actualiza el Cache de direcciones de hardware. Si el Cache
  est lleno, se elige una entrada al azar entre las usadas, y se utiliza
  esa. Si la entrada ya est en la lista, no se hace nada.
*)
Var
	i: integer;
    CacheEntry: Integer;
begin
	{ Si la direccin de Hardware es nula o es broadcast, la ignoramos }
    if Ethernet_Compare(EthernetAddress, ETH_BROADCASTADDRESS) then Exit;
    if Ethernet_Compare(EthernetAddress, ETH_NULLADDRESS) then Exit;
	{ Si la direccin IP es nula o es broadcast, la ignoramos }
    if (IPAddress = 0) then Exit;
    if (IPAddress = INADDR_BROADCAST) then Exit;
	{ Si la entrada ya est en el Cache, no hacemos nada }
	if GetEthernetAddress(IPAddress, EthernetAddress) then Exit;
    { Si hay lugar en el Cache, agregamos la entrada. Si no, elegimos }
	{ una entrada al azar.                                            }
    CacheEntry := Random(High(CacheList)) + 1;
    for i := Low(CacheList) to High(CacheList) do begin
        if (CacheList[i].Expires < GetMSCounter) then begin
        	CacheEntry := i;
            Break;
        end;
    end;
    { Actualizamos la entrada en cuestin. Es vlida por 2 minutos }
    CacheList[CacheEntry].IPAddress 	  := IPAddress;
    CacheList[CacheEntry].EthernetAddress := EthernetAddress;
    CacheList[CacheEntry].Expires         := GetMSCounter + 120000;
end;

procedure ProcessARP(Adapter: TNetworkAdapter;
  SourceEtherAddr: TEthernetAddress; Operation: WORD;
  PLEN: BYTE; SenderHA: TEthernetAddress; SenderPA: Pointer;
  TargetHA: TEthernetAddress; TargetPA: Pointer);
(*
  Recibimos un mensaje ARP de IP. Tenemos que ver si es una pregunta
  o una respuesta. Si es una pregunta para nosotros, la contestamos.
  En cualquier caso, anotamos las direcciones IP de los participantes,
  para actualizar nuestra tablita de IPs.
*)
Var
    AA: TNetworkAdapter;
    MyEtherAddr: TEthernetAddress;
begin
    { Debug }
(*	if Swap(Operation) = 1 then begin
	    WriteLn('  *ARP-IP* (El equipo ',
		  IP_AddrToString(PIPAddress(SenderPA)^), ' busca al equipo ',
		  IP_AddrToString(PIPAddress(TargetPA)^), ')');
    end;
    WriteLn(' Operation              : ', Swap(Operation));
    WriteLn(' Sender Hardware Address: ', Ethernet_AddrToString(SenderHA));
    WriteLn(' Target Hardware Address: ', Ethernet_AddrToString(TargetHA));
    WriteLn(' Sender IP Address      : ', IP_AddrToString(PIPAddress(SenderPA)^));
    WriteLn(' Target IP Address      : ', IP_AddrToString(PIPAddress(TargetPA)^));
*)
    { Revisamos si hay algo que configurar con el DHCP }
    if not CheckingDHCP then IP_CheckDHCP;

    { Actualizamos las direcciones de hardware de los participantes }
    SetEthernetAddress(PIPAddress(SenderPA)^, SenderHA);
    SetEthernetAddress(PIPAddress(TargetPA)^, TargetHA);

    { Si es una pregunta ARP para nosotros, la contestamos }
    if (Swap(Word(Operation)) = 1) then begin
        AA := GetAdapterFromAddress(PIPAddress(TargetPA)^);
        if AA <> nil then begin
            AA.GetAddress(MyEtherAddr);
			ARP_Send(Adapter, SourceEtherAddr, TYPELEN_IP, Swap(Word(2)), 4,
			  MyEtherAddr, TargetPA, SenderHA, SenderPA);
        end;
    end;
end;

Function GetHardwareAddress(TargetIP, LocalIP: TIPAddress;
  var EthernetAddress: TEthernetAddress): Boolean;
(*
  Queremos saber la direccin Ethernet de un equipo con un IP dado. Para
  eso, emitimos una peticin ARP a todos los equipos de nuestra red
  (broadcast) y esperamos una respuesta. Si al final de la espera, el IP
  solicitado est en nuestra tabla de IPs, lo hemos conseguido. De lo
  contrario, devolvemos FALSE.
*)
Var
    Timeout: Comp;
    Adapter: TNetworkAdapter;
    MyEthernetAddress: TEthernetAddress;
begin
	Result := False;
    { Revisamos para ver si la direccin Ethernet est en Cache }
    if GetEthernetAddress(TargetIP, EthernetAddress) then begin
        Result := True;
		Exit;
    end;
    Adapter := GetAdapterFromAddress(LocalIP);
    Adapter.GetAddress(MyEthernetAddress);
	{ Enviamos la solicitud ARP }
	if not ARP_Send(Adapter, ETH_BROADCASTADDRESS, TYPELEN_IP, Swap(Word(1)), 4,
      MyEthernetAddress, @LocalIP, ETH_NULLADDRESS, @TargetIP) then Exit;
    { Esperamos el resultado }
    Timeout := GetMSCounter + 500;              // 1/2 segundo
    While (GetMSCounter < Timeout) do begin
        NET_Work_Low;
        if GetEthernetAddress(TargetIP, EthernetAddress) then begin
            Result := True;
			Exit;
        end;
    end;
    { Timeout: Nadie contest nuestro ARP }
end;

{**************************************************************************}
{                                                                          }
{              Inicializacin y Deinicializacin de IP                     }
{                                                                          }
{**************************************************************************}
Function BindAdapter(Adapter: TNetworkAdapter): Boolean;
Var
    i: Integer;
    Bound: Boolean;
    P: PAddressEntry;
begin
	BindAdapter := False;
    { Check to see if the specified adapter is already bound              }
    Bound := False;
    for i := 0 to AddressList.Count - 1 do begin
        P := AddressList.At(i);
        if P^.Adapter = Adapter then begin
        	Bound := True;
        	Break;
        end;
    end;
    { If the adapter is not bound, we have to bind it now                 }
    if not Bound then begin
    	{ Bind to IP }
        if not Adapter.AddProtocol(TYPELEN_IP, @IP_ProcessPacket) then Exit;
        { Bind to ARP }
	    if not ARP_AddProcessor(Adapter, TYPELEN_IP, @ProcessARP) then begin
    	    Adapter.RemoveProtocol(TYPELEN_IP, @IP_ProcessPacket);
    		Exit;
	    end;
    end;
    { All Done }
    BindAdapter := True;
end;

Function IP_Init(Adapter: TNetworkAdapter; Config: TIPAdapterConfig): DWORD;
{ This function is intended to install the IP protocol over the given 	  }
{ adapter. If DHCP is specified in the configuration record, the supplied }
{ IP address, netmask and gateway are ignored, and they're obtained (if   }
{ possible) from a DHCP server.                                           }
{ If DHCP is not used, the Adapter is assigned the IP address, netmask,   }
{ and Gateway indicated by the corresponding fields in the configuration  }
{ record.                                                                 }
Var
    P: PAddressEntry;
begin
    IP_Init := 0;
    { Validamos los parmetros de entrada                                 }
    if not Config.DHCP and
      ((Config.IPAddress = 0) or (Config.NetMask = 0)) then Exit;
    { Bind this adapter to ARP and IP			    					  }
    if (Adapter <> nil) and not BindAdapter(Adapter) then Exit;
    { Registramos esta direccin IP en la lista de direcciones            }
    New(P);
    AddressList^.Insert(P);
    P^.Adapter := Adapter;
    P^.DHCPInUse := Config.DHCP;
    { Intentamos obtener una direccin IP por DHCP }
    if Config.DHCP then begin
        P^.DHCPInUse := True;
        P^.IPAddress := 0;
        P^.DHCPT1 := 0;
        P^.DHCPT2 := 0;
        P^.DHCPT3 := 0;
        { Intentamos obtener el IP ahora (puede fallar). Si falla no   }
        { importa: el DHCP intentar peridicamente obtener un IP para }
        { esta direccin.                                              }
        IP_CheckDHCP;
    end else begin
	    P^.IPAddress := Config.IPAddress;
        { Add a loopback-route for the new IP }
    	IP_AddRoute(P^.IPAddress, $FFFFFFFF, INADDR_LOOPBACK, INADDR_LOOPBACK, 0);
        { Add a route for the local network of the new adapter }
    	IP_AddRoute(P^.IPAddress and Config.NetMask, Config.NetMask, P^.IPAddress, P^.IPAddress, 0);
        { Add a default route for the Gateway, if there's one }
	    if Config.Gateway <> 0 then begin
    	    IP_AddRoute(0, 0, Config.Gateway, Config.IPAddress, 9999);
        end;
    end;
    { Listo }
    IP_Init := DWORD(P);
end;

Function IP_Deinit(IPHandle: DWORD): Boolean;
Var
    Index: Integer;
    Found: Boolean;
    AE: PAddressEntry;
    Adapter: TNetworkAdapter;
begin
    Found := False;
    AE := PAddressEntry(IPHandle);

    { Search for the indicated IPHandle }
    Index := AddressList^.IndexOf(AE);
    if Index = -1 then Exit;

    { if the Address was obtained via DHCP, we have to release it }
    if (AE^.DHCPInUse) and (AE^.IPAddress <> 0) then begin
        DHCP_Release(AE^.Adapter, AE^.IPAddress, AE^.DHCPServer);
    end;

    { Deallocate the Entry and remove it from the list }
    Adapter := AE^.Adapter;
    AddressList.Delete(AE);
    Dispose(AE);

    { If the adapter is no longer in use, we have to unbind it          }
    if Adapter <> nil then begin
        Found := False;
        for Index := 0 to AddressList.Count - 1 do begin
            AE := AddressList.At(Index);
            if (AE^.Adapter = Adapter) then begin
                Found := True;
                Break;
            end;
	    end;
        if not Found then begin
            Adapter.RemoveProtocol(TYPELEN_IP, @IP_ProcessPacket);
            ARP_RemoveProcessor(Adapter, TYPELEN_IP);
        end;
    end;

    { Done }
    IP_Deinit := True;
end;

{**************************************************************************}
{                                                                          }
{              Rutinas varias de conversin de direcciones IP              }
{                                                                          }
{**************************************************************************}
Function IP_AddrToString(IP: TIPAddress): String;

	function IStr(i: Integer): String;
    Var
    	s: String;
    begin
		Str(i, s);
        IStr := s;
    end;

Var
	Arr: packed Array[0..3] of byte;
begin
	Move(IP, Arr, 4);
    IP_AddrToString :=
      IStr(Arr[0]) + '.' + IStr(Arr[1]) + '.' +
      IStr(Arr[2]) + '.' + IStr(Arr[3]);
end;

Function IP_GetDHCPOption(Option: BYTE; Var Buffer; Var BuffLen: LongInt): Boolean;
Var
    i: Integer;
    AE: PAddressEntry;
begin
	{ Search for an adapter with DHCP installed }
    for i := 0 to AddressList^.Count - 1 do begin
        AE := PAddressEntry(AddressList^.At(i));
        if AE^.DHCPInUse then begin
            { Found it !. Get the requested option }
            Result := DHCP_GetInfo(AE^.Adapter, AE^.IPAddress,
              AE^.DHCPServer, Option, Buffer, BuffLen);
        	Exit;
        end;
    end;
end;

Function IP_StringToAddr(StrAddr: String): TIPAddress;
Var
    Part: String;
    IPAddress: TIPAddress;
    i, j, k, Code: Integer;
    Arr: Array[0..3] of BYTE;
begin
    IP_StringToAddr := 0;
	for i := 0 to 3 do begin
    	if i = 3 then begin
			Part := StrAddr;
        end else begin
	    	j := Pos('.', StrAddr);
            if j = 0 then Exit;
            Part := Copy(StrAddr, 1, j - 1);
            StrAddr := Copy(StrAddr, j + 1, 255);
        end;
        { Validamos esta parte de la direccin IP }
        Val(Part, k, Code);
        if (Code <> 0) or (k < 0) or (k > 255) then Exit;
        { Seguimos }
        Arr[i] := k;
    end;
    Move(Arr, IPAddress, 4);
    IP_StringToAddr := IPAddress;
end;

Function IP_CompareNetwork(IP1, IP2, NetMask: TIPAddress): Boolean;
{ Esta funcin devuelve True si las direcciones P1 y P2 estn en la }
{ misma subred, dada la mscara NetMask.                            }
begin
    { Enmascaramos el IP 1 }
    IP1 := IP1 and NetMask;
    { Enmascaramos el IP 2 }
    IP2 := IP2 and NetMask;
    { Comparamos los resultados }
    IP_CompareNetwork := (IP1 = IP2);
end;

{**************************************************************************}
{                                                                          }
{           Rutinas de procesamiento de datagramas IP entrantes            }
{                                                                          }
{**************************************************************************}
Function IP_Checksum(Data: Pointer; DataCnt: WORD; PIP: pointer; PIPCnt: WORD): WORD;
type
    PBYTE = ^BYTE;
var
    pw: ^WORD;
	loop: word;
	x, w: word;
    csum: DWORD;
Begin
	IP_Checksum := 0;
	csum := 0;
	if DataCnt = 0 then exit;
	loop := DataCnt shr 1;

	if pip <> nil then begin
		pw := pip;
		for x := 1 to (PipCnt div 2) do begin
			csum := csum + pw^;
			inc(pw);
		end;
    end;

    pw := data;
	for x := 1 to loop do begin
		csum := csum + pw^;
     	inc(pw);
	end;

    if (DataCnt mod 2 = 1) then begin
		w := PBYTE(pw)^;
		csum := csum + w;
	end;

	csum := (csum mod $10000) + (csum div $10000);
	csum := csum + (csum shr 16);
    IP_Checksum := Word(not csum);
end;

function GetPacketData(P: PIPHeader; var DataLen: Word): Pointer;
Var
	HLen: BYTE;
begin
	HLen := (P^.VersHLen and $0F) * 4;		{ En bytes }
    GetPacketData := PChar(@P^) + HLen;
    DataLen := Swap(Word(P^.TotalLength)) - HLen;
end;

procedure IP_ProcessDatagram(Protocol: BYTE; SourceIP, TargetIP: TIPAddress;
  Var Data; DataLen: WORD);
Var
	i: BYTE;
begin
    { Revisamos el Tipo de Protocolo, y Ejecutamos el Procesador que }
    { corresponda                                                    }
    for i := Low(Processors) to High(Processors) do begin
        if Processors[i].Protocol = Protocol then begin
            Processors[i].Routine(SourceIP, TargetIP, Data, DataLen);
        end;
    end;
end;

procedure IP_AddFragment(ID: WORD; Protocol: BYTE;
  SourceIP, TargetIP: TIPAddress; var Data; DataLen, Offset: WORD;
  Last: Boolean);
Var
	i, dg, dgfree, FragPos: integer;
begin
    { Nos aseguramos de que exista memoria para almacenar el nuevo fragmento }
    if MaxAvail < (DataLen + 1024) then Exit;
	{ Buscamos el Datagrama (si existe) }
	dg := 0;
    dgfree := 0;
    for i := Low(Datagrams) to High(Datagrams) do begin
		if (Datagrams[i].FragCount = 0) and (dgfree = 0) then dgfree := i;
        if (Datagrams[i].Id = Id) and (SourceIP = Datagrams[i].SourceIP) then begin
        	dg := i;
            break;
        end;
    end;
    if dg + dgfree = 0 then Exit;
    { Si es necesario, agregamos un nuevo datagrama }
    if (dg = 0) and (dgfree <> 0) then begin
        dg := dgfree;
        Datagrams[dg].SourceIP := SourceIP;
        Datagrams[dg].TargetIP := TargetIP;
        Datagrams[dg].Id := ID;
        Datagrams[dg].Protocol := Protocol;
		Datagrams[dg].TotalLength := 0;			{ an no lo sabemos }
        Datagrams[dg].ReceivedLength := 0;		{ an no se recibi nada }
        Datagrams[dg].FragCount := 0;
        Datagrams[dg].Timeout := GetMSCounter + 2000;	{ 2 segundos }
		FillChar(Datagrams[dg].Fragments, SizeOf(Datagrams[dg].Fragments), 0);
    end;
    { Buscamos la posicin del nuevo fragmento en el datagrama }
   	FragPos := 1;
   	for i := Datagrams[dg].FragCount downto 1 do begin
		if (Offset > Datagrams[dg].Fragments[i].Offset) then begin
        	FragPos := i + 1;
            Break;
        end;
	end;
    { Corremos los fragmentos para agregar el nuevo fragmento }
    Inc(Datagrams[dg].FragCount);
    for i := Datagrams[dg].FragCount downto FragPos + 1 do begin
        Datagrams[dg].Fragments[i] := Datagrams[dg].Fragments[i - 1];
    end;
    { Agregamos el nuevo fragmento }
    Datagrams[dg].Fragments[FragPos].Offset := Offset;
    GetMem(Datagrams[dg].Fragments[FragPos].Data, DataLen);
    Move(Data, Datagrams[dg].Fragments[FragPos].Data^, Datalen);
    Datagrams[dg].Fragments[FragPos].DataLen := DataLen;
	{ Actualizamos los campos del datagrama }
	Inc(Datagrams[dg].ReceivedLength, DataLen);
    if Last then Datagrams[dg].TotalLength := (Offset * 8) + DataLen;
end;

procedure IP_CheckDatagrams;
Var
	i, f: Integer;
    PData, PP: PChar;
begin
    { Procesamos los datagramas completos }
    for i := Low(Datagrams) to High(Datagrams) do begin
        if (Datagrams[i].FragCount > 0) and
		  (Datagrams[i].ReceivedLength = Datagrams[i].TotalLength) then begin
            { Armamos el Datagrama final }
            if GetMemCheck(Pointer(PData), Datagrams[i].TotalLength) then begin
	            PP := PData;
    	        for f := 1 to Datagrams[i].FragCount do begin
        	        Move(Datagrams[i].Fragments[f].Data^, PP^,
					  Datagrams[i].Fragments[f].DataLen);
	                FreeMem(Datagrams[i].Fragments[f].Data,
					  Datagrams[i].Fragments[f].DataLen);
        	        Inc(PP, Datagrams[i].Fragments[f].DataLen);
            	end;
	            { Ahora que tenemos el datagrama armado, lo procesamos }
    	        IP_ProcessDatagram(Datagrams[i].Protocol,
				  Datagrams[i].SourceIP, Datagrams[i].TargetIP, PData^,
            	  Datagrams[i].TotalLength);
	            { Eliminanos el datagrama recin creado }
    	        FreeMem(PData, Datagrams[i].TotalLength);
        	    { Eliminamos el datagrama usado }
            	FillChar(Datagrams[i], SizeOf(Datagrams[i]), 0);
            end;
        end;
    end;
    { Eliminanos los datagramas incompletos }
    for i := Low(Datagrams) to High(Datagrams) do begin
        if (Datagrams[i].Timeout < GetMSCounter) then begin
        	{ Liberamos la memoria de los fragmentos }
            for f := 1 to Datagrams[i].FragCount do begin
                FreeMem(Datagrams[i].Fragments[f].Data,
				  Datagrams[i].Fragments[f].DataLen);
            end;
            { Eliminamos el datagrama usado }
            FillChar(Datagrams[i], SizeOf(Datagrams[i]), 0);
        end;
    end;
end;

procedure IP_ProcessFragment(ID: WORD; Protocol: BYTE;
  SourceIP, TargetIP: TIPAddress; Var Data; DataLen, Offset: WORD;
  Last: Boolean);
begin
	IP_AddFragment(ID, Protocol, SourceIP, TargetIP, Data, DataLen, Offset, Last);
end;

procedure IP_ProcessPacket(Adapter: TNetworkAdapter; Packet: PEthernetPacket);
Var
    Flags: BYTE;
	PData: PChar;
	DataLen: Word;
    Pkt: PIPHeader;
//	FragmentLen: WORD
    FragmentOffset: WORD;
begin
    Pkt := @Packet^.Data;

    { Revisamos si hay algo que configurar con el DHCP }
    if not CheckingDHCP then IP_CheckDHCP;

    { Revisamos la versin de IP. Slo aceptamos la versin 4 }
    if (Pkt^.VersHLen shr 4) <> 4 then Exit;

    { Debug }
//	WriteLn('*IP* (del equipo ', IP_AddrToString(Pkt^.SourceIP),
//	  ' al ', IP_AddrToString(Pkt^.TargetIP), '). Orig. Eth: ',
//	  Ethernet_AddrToString(Packet^.Header.SourceAddr));

    { Si el paquete no es para nosotros, lo ignoramos }
    if (Pkt^.TargetIP <> INADDR_BROADCAST) and
      (GetAdapterFromAddress(Pkt^.TargetIP) = nil) then Exit;

    { Tomamos los Datos del Paquete }
    Flags := Pkt^.FlagsFrag1 shr 5;
    FragmentOffset := Pkt^.Frag2 + (Pkt^.FlagsFrag1 and $1F) * $100;
//	FragmentLen := Swap(Word(Pkt^.TotalLength));
	PData := GetPacketData(Pkt, DataLen);

    { Debug }
(*  WriteLn(' HLEN:            ', Pkt^.VersHLen and $0F);
	WriteLn(' Protocol:        ', Pkt^.Protocol);
	WriteLn(' Identification:  ', Pkt^.Id);
	WriteLn(' Flags:           ', Flags);
	WriteLn(' Fragment Offset: ', FragmentOffset);
	WriteLn(' Fragment Length: ', FragmentLen);
	WriteLn(' Checksum:        ', IntToHex(Pkt^.Checksum, 4));
	Pkt^.Checksum := 0;
	WriteLn(' Checksum Recal.: ', IntToHex(IP_Checksum(@Pkt^, SizeOf(TIPHeader), nil, 0), 4));
*)

    { Cacheamos la direccin IP del que nos enva el paquete }
	SetEthernetAddress(Pkt^.SourceIP, Packet^.Header.SourceAddr);
    { Si este paquete es todo el datagrama, lo procesamos directamente }
    if ((Flags and 1) = 0) and (FragmentOffset = 0) then begin
        IP_ProcessDatagram(Pkt^.Protocol, Pkt^.SourceIP, Pkt^.TargetIP,
		  PData^, DataLen);
    end else begin
    	{ El paquete es slo un fragmento del datagrama. Lo agregamos a }
        { lista de fragmentos.                                          }
	    IP_CheckDatagrams;
        IP_ProcessFragment(Pkt^.Id, Pkt^.Protocol, Pkt^.SourceIP,
		  Pkt^.TargetIP, PData^, DataLen, FragmentOffset, (Flags and 1) = 0);
	    IP_CheckDatagrams;
    end;
end;

Function IP_AddProcessor(Protocol: BYTE; Routine: TIP_Routine): Boolean;
Var
	i: integer;
begin
	IP_AddProcessor := False;
    IP_RemoveProcessor(Protocol);
	for i := Low(Processors) to High(Processors) do begin
    	if (Processors[i].Routine = nil) then begin
        	Processors[i].Protocol := Protocol;
            Processors[i].Routine := Routine;
        	IP_AddProcessor := True;
            Exit;
        end;
    end;
end;

function IP_RemoveProcessor(Protocol: BYTE): Boolean;
Var
	i: integer;
begin
	IP_RemoveProcessor := False;
	for i := Low(Processors) to High(Processors) do begin
    	if (Processors[i].Protocol = Protocol) then begin
        	Processors[i].Protocol := 0;
            Processors[i].Routine := nil;
        	IP_RemoveProcessor := True;
            Exit;
        end;
    end;
end;

{**************************************************************************}
{                                                                          }
{                    Rutinas de envo de datagramas IP                     }
{                                                                          }
{**************************************************************************}
Function IP_AddRoute(Destination, Mask, Gateway, LocalIP: TIPAddress; Cost: Integer): Boolean;
Var
    P: PRouteEntry;
	i, Index: Integer;
begin
    Result := False;
    { Primero buscamos en qu lugar insertar la nueva ruta }
    Index := -1;
    for i := 0 to RouteList^.Count - 1 do begin
        P := RouteList^.At(i);
        if P^.Cost > Cost then begin
            Index := i;
            break;
        end;
    end;
    { Agregamos la ruta a la lista }
    New(P);
    P^.Destination := Destination;
    P^.Mask       := Mask;
    P^.LocalIP    := LocalIP;
    P^.Gateway    := Gateway;
    P^.Cost       := Cost;
    if Index = -1 then RouteList^.Insert(P)
      else RouteList^.AtInsert(Index, P);
    { Listo }
    Result := True;
end;

Procedure IP_PrintRoute;
Var
    i: Integer;
    P: PRouteEntry;
begin
    WriteLn('Destino         Mscara         Gateway         IP Local        Costo');
    WriteLn('--------------- --------------- --------------- --------------- -----');
    for i := 0 to RouteList^.Count - 1 do begin
        P := RouteList^.At(i);
        WriteLn(
          PadR(IP_AddrToString(P^.Destination), 16, ' '),
          PadR(IP_AddrToString(P^.Mask), 16, ' '),
          PadR(IP_AddrToString(P^.Gateway), 16, ' '),
          PadR(IP_AddrToString(P^.LocalIP), 16, ' '),
          P^.Cost);
    end;
end;

Function IP_GetRouteForTarget(TargetIP: TIPAddress; var LocalIP, Gateway: TIPAddress): Boolean;
Var
    i: Integer;
    RE: PRouteEntry;
begin
    { Si no tenemos ninguna direccin IP configurada, nos vamos }
    if RouteList^.Count = 0 then begin
        IP_GetRouteForTarget := False;
    	Exit;
    end;
    { Primero recorremos la tabla de ruteo para encontrar una ruta }
    { apropiada }
    for i := 0 to RouteList^.Count - 1 do begin
        RE  := RouteList^.At(i);
        if (RE^.Destination <> 0) and (IP_CompareNetwork(RE^.Destination,
          TargetIP, RE^.Mask)) then begin
            LocalIP := RE^.LocalIP;
            Gateway := RE^.Gateway;
            IP_GetRouteForTarget := True;
            Exit;
        end;
    end;
    { No encontramos una ruta vlida. Buscamos entonces un gateway por }
    { default }
    for i := 0 to RouteList^.Count - 1 do begin
        RE  := RouteList^.At(i);
        if (RE^.Destination = 0) then begin
            LocalIP := RE^.LocalIP;
            Gateway := RE^.Gateway;
            IP_GetRouteForTarget := True;
            Exit;
        end;
    end;
    { No hay una ruta adecuada, ni tampoco un gateway default }
    IP_GetRouteForTarget := False;
end;

function IP_Snd(
  Adapter: TNetworkAdapter;
  DestEth: TEthernetAddress;
  Protocol: BYTE;
  LocalIP, TargetIP: TIPAddress;
  Var Data;
  DataLen: Integer
  ): Integer;
const
	IP_MAXDATA = 1480;
Type
	TTrama = packed record
    	EthHeader: TEthernetHeader;
        IPHeader: TIPHeader;
        IPData: packed Array[0..IP_MAXDATA - 1] of BYTE;
    end;
Var
    PData: PChar;
    Trama: TTrama;
    FragLen: WORD;
    FragOffs: WORD;
    LocalEth: TEthernetAddress;
begin
    Adapter.GetAddress(LocalEth);
    { Debug }
(*  WriteLn('Enviando un datagrama desde ', Ethernet_AddrToString(LocalEth),
      ' (', IP_AddrToString(LocalIP), ') hacia ', Ethernet_AddrToString(DestEth),
      ' (', IP_AddrToString(TargetIP), ') con ', DataLen, ' bytes.');
*)
    { Armamos los datos fijos de la trama Ethernet, que se repiten en   }
    { Todos los fragmentos.                                             }
    Trama.EthHeader.TargetAddr := DestEth;
    Trama.EthHeader.SourceAddr := LocalEth;
    Trama.EthHeader.TypeLen    := TYPELEN_IP;
    { Armamos los datos fijos de la cabecera IP, que se repiten en      }
    { todos los fragmentos.                                             }
	Trama.IPHeader.VersHLen 	:= $45;		{ VERS = 4; HLEN = 5 }
	Trama.IPHeader.ServiceType 	:= 0;
	Trama.IPHeader.ID 			:= Random($FFFF) + 1;
	Trama.IPHeader.LifeTime		:= 128;
    Trama.IPHeader.Protocol     := Protocol;
    Trama.IPHeader.SourceIP     := LocalIP;
    Trama.IPHeader.TargetIP     := TargetIP;
	{ Componemos los fragmentos, metindolos directamente en la trama   }
    { Ethernet.                                                         }
    PData := @Data;
    FragOffs := 0;
    While DataLen > 0 do begin
    	{ Modificamos los flags y dems de este fragmento.              }
        Trama.IPHeader.FlagsFrag1 := FragOffs shr 8;
        Trama.IPHeader.Frag2      := FragOffs and $FF;
        if DataLen > IP_MAXDATA then begin
        	Trama.IPHeader.FlagsFrag1 :=
			  Trama.IPHeader.FlagsFrag1 or 32;		   { More fragments }
        	FragLen := IP_MAXDATA;
        end else begin
			FragLen := DataLen;
        end;
		Trama.IPHeader.TotalLength 	:= Swap(Word(FragLen + SizeOf(TIPHeader)));
		Trama.IPHeader.Checksum 	:= 0;
		Trama.IPHeader.Checksum		:= IP_Checksum(@Trama.IPHeader, SizeOf(TIPHeader), nil, 0);
        { Agregamos los datos                                           }
        Move(PData^, Trama.IPData, FragLen);
        { Enviamos el Paquete Ethernet }
        Adapter.SendPkt(@Trama, SizeOf(Trama));
        { Nos preparamos para el Fragmento siguiente }
        Inc(FragOffs, FragLen div 8);
        Inc(PData, FragLen);
		Dec(DataLen, FragLen);
    end;
    IP_Snd := 0;
end;


function IP_Send(TargetIP: TIPAddress; Protocol: BYTE; Var Data;
  DataLen: WORD): LongInt;

    Function IsBroadcast(TargetIP: TIPAddress): Boolean;
    begin
    	Result := (Pos('255', IP_AddrToString(TargetIP)) > 0);
    end;

Var
    i: Integer;
    P: PAddressEntry;
    Adp: TNetworkAdapter;
    DestEth: TEthernetAddress;
    LocalIP, Gateway: TIPAddress;
begin
    { Si no tenemos ningn IP, no hay nada que hacer }
    if AddressList^.Count = 0 then begin
        IP_Send := WSAENETUNREACH;
        Exit;
    end;
    { Si es una direccin de broadcast, hay que salir por los IPs         }
    { que correspondan (que pueden ser ms de uno). Si no es un broadcast }
    { tenemos que elegir un adaptador/gateway y sacar el paquete por l.  }
    if IsBroadcast(TargetIP) then begin
        { The destination IP is the Broadcast Address. We have to   }
        { cycle through all the addresses                           }
        for i := 0 to AddressList^.Count - 1 do begin
            P := AddressList^.At(i);
            if P^.Adapter <> nil then begin
                IP_Snd(P^.Adapter, ETH_BROADCASTADDRESS, Protocol,
                  P^.IPAddress, TargetIP, Data, DataLen);
            end;
        end;
        IP_Send := 0;
    end else begin
        { El destino no es un broadcast. Buscamos una ruta para sacar el  }
        { datagrama. 													  }
        if not IP_GetRouteForTarget(TargetIP, LocalIP, Gateway) then begin
   	        IP_Send := WSAENETUNREACH;
   		    Exit;
        end;
        { Si el IP local resulta ser el LOOPBACK, procesamos el datagrama }
        { en forma interna                                                }
        if LocalIP = INADDR_LOOPBACK then begin
            IP_ProcessDatagram(Protocol, INADDR_LOOPBACK, INADDR_LOOPBACK,
              Data, DataLen);
            IP_Send := 0;
            Exit;
        end;
        { If we must go through a gateway in order to reach the     }
        { destination, we have to direct our datagram to the        }
        { hardware address of the gateway. Otherwise, the datagram  }
        { is directed to the hardware address of the destination    }
        { itself.                                                   }
        Adp := GetAdapterFromAddress(LocalIP);
        if (LocalIP = Gateway) then begin
            if not GetHardwareAddress(TargetIP, LocalIP, DestEth) then begin
                { We couldn't determine the destination's hardware  }
                { address. }
                IP_Send := WSAEHOSTDOWN;
				Exit;
		    end;
		end else begin
            if not GetHardwareAddress(Gateway, LocalIP, DestEth) then begin
            	{ We couldn't determine the gateway's hardware      }
                { address.                                          }
                IP_Send := WSAEHOSTDOWN;
                Exit;
            end;
        end;
        IP_Send := IP_Snd(Adp, DestEth, Protocol, LocalIP, TargetIP,
          Data, DataLen);
    end;
end;

Var
    RE: PRouteEntry;
Initialization
    //
    CheckingDHCP := False;
	// Address List
    AddressList := New(PCollection, Init(1, 1));
    // Route List
    RouteList := New(PCollection, Init(10, 1));
    //
    FillChar(CacheList, SizeOf(CacheList), 0);
    FillChar(Processors, SizeOf(Processors), 0);
    FillChar(Datagrams, SizeOf(Datagrams), 0);
    // Inicializamos ICMP
    ICMP_Init;
    // Inicializamos UDP
    UDP_Init;
    // Inicializamos el TCP
    TCP_Init;
    // Armamos la tabla de ruteo bsica, conteniendo slo el ruteo
    // hacia el LOOPBACK.
    IP_AddRoute(IP_StringToAddr('127.0.0.0'), IP_StringToAddr('255.0.0.0'),
      INADDR_LOOPBACK, INADDR_LOOPBACK, 0);
finalization
    { Delete Addresses & Address List }
    While AddressList^.Count > 0 do begin
        IP_Deinit(DWORD(AddressList^.At(0)));
    end;
    Dispose(AddressList, Done);

    { Delete Routes & Route Table }
    While RouteList^.Count > 0 do begin
        RE := RouteList^.At(0);
        Dispose(RE);
        RouteList.AtDelete(0);
    end;
    Dispose(RouteList, Done);
end.
