Unit TCP;

interface

Uses
    Network, Crt, IP, Objects, Sockets, DosSock, Util, SysUtils;

{ Initialisation and deinitialisation }
Function TCP_Init: Boolean;
Function TCP_Deinit: Boolean;

{ Creation, connection & disconnection }
Function TCP_Socket			(Domain, SocketType, Protocol: Longint): Pointer;
Function TCP_Close          (Sock: Pointer): LongInt;
Function TCP_Shutdown       (Sock: Pointer; How: LongInt): LongInt;
Function TCP_Bind			(Sock: Pointer; Var Addr; AddrLen: Longint): Boolean;
Function TCP_Listen			(Sock: Pointer; MaxConnect: Longint): Boolean;
Function TCP_Accept     	(Sock: Pointer; Var Addr; Var Addrlen: Longint): Pointer;
Function TCP_Connect    	(Sock: Pointer; Var Addr; Addrlen: Longint): Boolean;

{ Reception }
Function TCP_Recv			(Sock: Pointer; Var Addr; AddrLen, Flags: Longint): Longint;

{ Sending }
Function TCP_Send       	(Sock: Pointer; Var Addr; AddrLen, Flags: Longint): Longint;

{ Various }
Function TCP_GetLastError: LongInt;
Function TCP_ioctlsocket	 (Sock: Pointer; Option: LongInt; d: Pointer): LongInt;
Function TCP_GetSocketName	 (Sock: Pointer; Var Addr; Var Addrlen: Longint): Longint;
Function TCP_GetPeerName     (Sock: Pointer; Var Addr; Var Addrlen: Longint): Longint;
Function TCP_SetSocketOptions(Sock: Pointer; Level, OptName: Longint; Var OptVal; OptLen: Longint): Longint;
Function TCP_GetSocketOptions(Sock: Pointer; Level, OptName: Longint; Var OptVal; Var OptLen: Longint): Longint;

implementation

const
    { Code Bits Flags }
    CB_END 		=  1;
    CB_SYN 		=  2;
    CB_RST 		=  4;
    CB_PSH 		=  8;
	CB_ACK 		= 16;
    CB_OOB 		= 32;
    { Other parameters }
    MAX_PENDING = 5;
    MTU 		= 536;
    MAX_WINDOW  = 2 * MTU;

Type
	PTCPHeader = ^TTCPHeader;
	TTCPHeader = packed record
        SourcePort: WORD;
        TargetPort: WORD;
        SequenceNumber: DWORD;
        AckNumber: DWORD;
        HLen: BYTE;
        CodeBits: BYTE;
        Window: WORD;
        Checksum: WORD;
        UrgentPointer: WORD;
    end;

	TPseudoHeader = packed record
        SourceIP: TIPAddress;
        TargetIP: TIPAddress;
        Cero: BYTE;
        Protocol: BYTE;
        TCPLen: WORD;
    end;

    TPendingEntry = record
    	LocalIP: TIPAddress;
    	RemoteIP: TIPAddress;
        RemotePort: WORD;
        SequenceNumber: DWORD;
    end;

    TTCPStatus = (tcpClosed, tcpListen, tcpSYNRcvd, tcpConnected,
	  tcpLocalEndWait1, tcpLocalEndWait2, tcpRemoteEndWait1,
      tcpConnecting1);

	PSocketRec = ^TSocketRec;
    TSocketRec = record
        { Socket Status }
    	Status: TTCPStatus;
        { Local & remote connection information }
        LocalIP: TIPAddress;
        LocalPort: WORD;
        RemoteIP: TIPAddress;
		RemotePort: WORD;
        { Flow Control }
        NextByteToSend: DWORD;
        LastByteConfirmed: DWORD;
        LastByteReceived: DWORD;
        RemoteWindow: WORD;
        NextACKTimeout: Comp;
        { Buffers }
        DataBuffer: AnsiString;
        { Pending connections (for listening) }
		PendingCount: BYTE;
        Pendings: Array[1..MAX_PENDING] of TPendingEntry;
        { Options }
        ConnTimeout: DWORD;         // in milliseconds
    end;


Var
	LastError: LongInt;
    SockList: PCollection;


function TCP_Checksum(SourceIP, TargetIP: TIPAddress;
  PData: PChar; Len: Word): WORD;
var
    PH: TPseudoHeader;
begin
    { Set-up the psueudo-header }
    FillChar(PH, SizeOf(PH), 0);
    PH.SourceIP := SourceIP;
    PH.TargetIP := TargetIP;
	PH.TCPLen   := Swap(Word(Len));
	PH.Cero     := 0;
    PH.Protocol := IPPROTO_TCP;
    { Calculate the checksum }
    TCP_Checksum := IP_Checksum(PData, Len, @PH, SizeOf(PH));
end;

Function GetConnectedSocket(LocalIP, RemoteIP: TIPAddress;
  LocalPort, RemotePort: WORD): PSocketRec;
Var
    i: Integer;
    PSock: PSocketRec;
begin
    for i := 0 to SockList^.Count - 1 do begin
        PSock := PSocketRec(SockList^.At(i));
        if not (PSock^.Status in [tcpClosed, tcpListen])
          and (PSock^.RemoteIP = RemoteIP)
          and (PSock^.RemotePort = RemotePort)
          and ((PSock^.LocalIP = LocalIP) or (PSock^.LocalIP = INADDR_ANY))
          and (PSock^.LocalPort = LocalPort) then begin
            GetConnectedSocket := PSock;
            Exit;
        end;
    end;
    GetConnectedSocket := nil;
end;

Function GetAnySocket(LocalIP: TIPAddress; LocalPort: WORD): PSocketRec;
Var
	i: Integer;
    PSock: PSocketRec;
begin
    for i := 0 to SockList^.Count - 1 do begin
        PSock := PSocketRec(SockList^.At(i));
        if ((PSock^.LocalIP = LocalIP) or (PSock^.LocalIP = INADDR_ANY))
          and (PSock^.LocalPort = LocalPort) then begin
            GetAnySocket := PSock;
            Exit;
        end;
    end;
    GetAnySocket := nil;
end;

Function GetFreePort: Integer;
Var
	Resu: WORD;
    i, Prt: Integer;
begin
	for Prt := 1024 to 8192 do begin
    	Resu := Prt;
        for i := 0 to SockList^.Count - 1 do begin
            if PSocketRec(SockList^.At(i))^.LocalPort = Prt then begin
            	Resu := 0;
                break;
			end;
        end;
        if Resu <> 0 then begin
        	GetFreePort := Swap(Word(Resu));
            Exit;
        end;
    end;
	GetFreePort := 0;
end;

function Send(
  Sock: PSocketRec;
  SourceIP, TargetIP: TIPAddress;
  SourcePort, TargetPort: WORD;
  CodeBits: BYTE;
  PData: Pointer; DataLen: WORD): Boolean;
Var
    Hdr: PTCPHeader;
    WinSize: Integer;
begin
    Send := False;
    if not GetMemCheck(Pointer(Hdr), SizeOf(TTCPHeader) + DataLen) then Exit;
    { Set up the header }
    FillChar(Hdr^, SizeOf(Hdr^), 0);
    WinSize             := MAX_WINDOW - Length(Sock^.DataBuffer);
    if WinSize < 0 then WinSize := 0;
    Hdr^.SourcePort     := SourcePort;
	Hdr^.TargetPort     := TargetPort;
    Hdr^.ACKNumber      := SwapDWORD(Sock^.LastByteReceived + 1);
    Hdr^.SequenceNumber := SwapDWORD(Sock^.NextByteToSend);
    Hdr^.HLen           := (SizeOf(TTCPHeader) div 4) shl 4;
    Hdr^.CodeBits       := CodeBits;
    Hdr^.Window         := Swap(Word(WinSize));
    { Add data }
    Move(PData^, PChar(Hdr)[SizeOf(TTCPHeader)], DataLen);
    { Calculate checksum }
    Hdr^.Checksum       := 0;
    Hdr^.Checksum       := TCP_Checksum(SourceIP, TargetIP, Pointer(Hdr),
      SizeOf(TTCPHeader) + DataLen);
    { Send the datagram }
    Send := IP_Send(TargetIP, IPPROTO_TCP, Hdr^, SizeOf(TTCPHeader) + DataLen) = 0;
{    WriteLn('***** Se envi el paquete    ', SwapLong(Hdr^.SequenceNumber));
    WriteLn('***** Se confirm el paquete ', Sockets[Socket].LastByteReceived);
}
    { Free Buffer }
    FreeMem(Hdr, SizeOf(TTCPHeader) + DataLen);
    { if data was sent, increase the Secuence Number }
    Inc(Sock^.NextByteToSend, DataLen);
end;

procedure ProcessIP(SourceIP, TargetIP: TIPAddress;
  Var Data; DataLen: WORD);
Var
    CodeBits: BYTE;
    Hdr: PTCPHeader;
    Sock: PSocketRec;
    SourcePort, TargetPort: WORD;
    HLen, nPend, AntLen: Integer;
    SequenceNumber, ExpectedByte: DWORD;
begin
    Hdr := @Data;
    if Hdr^.TargetPort = 0 then Exit;
    {}
    HLen := BYTE(Hdr^.HLen shr 4) * 4;
    CodeBits := Hdr^.CodeBits and 63;
	SourcePort := Hdr^.SourcePort;
	TargetPort := Hdr^.TargetPort;
    SequenceNumber := SwapDWORD(Hdr^.SequenceNumber);
    { Debug }
//	WriteLn('  *TCP*');
//	WriteLn('   Source Port    : ', Swap(SourcePort));
//	WriteLn('   Target Port    : ', Swap(TargetPort));
//	WriteLn('   HLen           : ', HLen);
//	WriteLn('   DataLen        : ', DataLen - HLen);
//  WriteLn('   Code Bits      : ', Hdr^.CodeBits);
//	WriteLn('   Sequence Number: ', SequenceNumber);
//	WriteLn('   ACK Number     : ', SwapDWORD(Hdr^.AckNumber));
//	WriteLn('   Window         : ', Swap(Hdr^.Window));
//  WriteLn('   Checksum       : ', IntToHex(Hdr^.Checksum, 4));
//  Hdr^.Checksum := 0;
//  WriteLn('   Checksum Recalc: ',
//    IntToHex(TCP_Checksum(SourceIP, TargetIP, @Data, DataLen), 4));
//  WriteLn('   Urgent Pointer : ', Swap(Hdr^.UrgentPointer));

    { Se recibi un datagrama TCP. Primero que nada hay que ver si existe  }
    { algn socket conectado listo para recibir este datagrama.            }
    Sock := GetConnectedSocket(TargetIP, SourceIP, TargetPort, SourcePort);

    { Si lo anterior no result, buscamos algn socket no conectado, pero  }
    { listo para recibir este datagrama.                                   }
    if Sock = nil then Sock := GetAnySocket(TargetIP, TargetPort);

    { Verificamos el nmero de Secuencia del Emisor. Si no se corresponde  }
    { con lo que debera, nos vamos, ignorando el Paquete. La nica        }
    { excepcin a esto es el paquete de SYN, que es el primero.            }
    if (Sock <> nil) and not (Sock^.Status in [tcpListen, tcpConnecting1]) then begin
        ExpectedByte := Sock^.LastByteReceived + 1;
        if (SequenceNumber = ExpectedByte) or (SequenceNumber = 0) then begin
            Inc(Sock^.LastByteReceived, DWORD(DataLen - HLen));
{			WriteLn('Se acept el paquete ', SequenceNumber);}
        end else begin
{        	WriteLn('Se RECHAZO el Paquete ', SequenceNumber,
              ' (se esperaba el ', ExpectedByte, ')');
}
			Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
			  CB_ACK, nil, 0);
            Exit;
        end;
    end;

    { Procesamiento para Code Bits = CB_SYN (Conexin entrante)            }
	if (CodeBits and CB_SYN) = CB_SYN then begin
    	{ Si hay un socket en modo listen para este paquete, aceptamos     }
		{ la conexin. De lo contrario, la rechazamos.                     }
        if Sock <> nil then case Sock^.Status of
			tcpListen:
                if (Sock^.PendingCount < MAX_PENDING) then begin
                    { Se acepta la conexin. La ponemos en la cola de      }
                    { conexiones para ser aceptadas.                       }
                    Inc(Sock^.PendingCount);
                    nPend := Sock^.PendingCount;
                    Sock^.Pendings[nPend].LocalIP := TargetIP;
                    Sock^.Pendings[nPend].RemoteIP := SourceIP;
                    Sock^.Pendings[nPend].RemotePort := SourcePort;
                    Sock^.Pendings[nPend].SequenceNumber := SequenceNumber;
                end else begin
			        Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
					  CB_RST, nil, 0);
                end;
            tcpConnecting1:
            	begin
                    Sock^.LastByteReceived := SequenceNumber;
			        Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
					  CB_ACK, nil, 0);
                    Sock^.Status := tcpConnected;
                end;
	        else
				begin
			        Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
					  CB_RST, nil, 0);
                end;
        end;
    end;

    { Procesamiento para CB_END                                            }
    if (CodeBits and CB_END) = CB_END then begin
//    	WriteLn('END!!!!!!!!!!!!!!!');
    	{ Si no es para nadie, nos vamos                                   }
        if Sock <> nil then case Sock^.Status of
            tcpLocalEndWait1, tcpLocalEndWait2:
            	begin
                    Inc(Sock^.NextByteToSend);
                    Inc(Sock^.LastByteReceived);
			        Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
					  CB_ACK, nil, 0);
//             		WriteLn('Socket cerrado localmente !');
                    Sock^.Status := tcpClosed;
                end;
            tcpConnected:
            	begin
                    Inc(Sock^.LastByteReceived);
			        Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
					  CB_ACK, nil, 0);
			        Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
					  CB_END or CB_ACK, nil, 0);
                    Sock^.Status := tcpRemoteEndWait1;
                end;
        end;
    end;

    { Procesamiento para CB_RST                                            }
    if (CodeBits and CB_RST) = CB_RST then begin
//      WriteLn('Reset ! ', Swap(SourcePort), ' ', Swap(TargetPort));
    	{ Si no es para nadie, nos vamos                                   }
        if (Sock <> nil) and not (Sock^.Status in [tcpListen]) then begin
            Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
              CB_RST or CB_ACK, nil, 0);
            Sock^.Status := tcpClosed;
        end;
    end;

    { Procesamiento de ACKs. Bsicamente no hacemos ms que almacenar el   }
    { Nmero de Confirmacin en el Socket elegido.                         }
    if (CodeBits and CB_ACK) = CB_ACK then begin
        if Sock <> nil then begin
            case Sock^.Status of
				tcpSynRcvd:
                    Sock^.Status := tcpConnected;
        		tcpLocalEndWait1:
                    Sock^.Status := tcpLocalEndWait2;
            	tcpRemoteEndWait1:
                	begin
//                      WriteLn('Socket cerrado remotamente !');
                        Sock^.Status := tcpClosed;
                    end;
                tcpConnecting1:
                    begin
                        Sock^.LastByteReceived := SequenceNumber;
                        Sock^.NextByteToSend := SwapDWORD(Hdr^.ACKNumber);
				        Send(Sock, TargetIP, SourceIP, TargetPort,
                          SourcePort, CB_RST or CB_ACK, nil, 0);
                    end;
            end;
            Sock^.RemoteWindow := Swap(Hdr^.Window);
            Sock^.LastByteConfirmed := SwapDWORD(Hdr^.ACKNumber);
        end;
    end;

    { Si llegaron datos, los confirmamos y los almacenamos el el Buffer   }
	{ del Socket.                                                         }
    DataLen := DataLen - HLen;
    if (Sock <> nil) and (DataLen > 0) then begin
        { Agrandamos el Buffer de Datos del Socket }
        AntLen := Length(Sock^.DataBuffer);
        SetLength(Sock^.DataBuffer, Length(Sock^.DataBuffer) + DataLen);
        { Copiamos los Nuevos Datos }
        Move(PChar(@Data)[HLen], Sock^.DataBuffer[AntLen + 1], DataLen);
        { Confirmamos los datos recin recibidos. }
        Sock^.NextACKTimeout := GetMSCounter + 300;
{       Send(Sock, TargetIP, SourceIP, TargetPort, SourcePort,
          CB_ACK, nil, 0);
		WriteLn('Confirmado: ', DataLen);}
    end;
end;

Function TCP_Init: Boolean;
Var
	PS: TProtocolStruct;
begin
	{ Add the TCP processor to de IP Stack }
    Result := IP_AddProcessor(IPPROTO_TCP, @ProcessIP);
    {}
    PS.Domain     			:= PF_INET;
    PS.SocketType           := SOCK_STREAM;
    PS.Protocol   			:= IPPROTO_IP;
    PS.FNSocket             := @TCP_Socket;
    PS.FNClose              := @TCP_Close;
    PS.FNGetLastError   	:= @TCP_GetLastError;
    PS.FNSend           	:= @TCP_Send;
    PS.FNRecv       	    := @TCP_Recv;
    PS.FNBind   	        := @TCP_Bind;
    PS.FNListen	    	    := @TCP_Listen;
    PS.FNAccept         	:= @TCP_Accept;
    PS.FNConnect    	    := @TCP_Connect;
    PS.FNShutdown           := @TCP_Shutdown;
    PS.FNGetSocketName      := @TCP_GetSocketName;
    PS.FNGetPeerName        := @TCP_GetPeerName;
    PS.FNSetSocketOptions   := @TCP_SetSocketOptions;
    PS.FNGetSocketOptions   := nil; //@TCP_GetSocketOptions;
    PS.FNioctlsocket        := @TCP_ioctlsocket;
    Result := Result and AddProtocol(PS);
    PS.Protocol             := IPPROTO_TCP;
    Result := Result and AddProtocol(PS);
end;

Function TCP_GetLastError: LongInt;
begin
	TCP_GetLastError := LastError;
end;

Function TCP_Deinit: Boolean;
begin
	TCP_Deinit := IP_RemoveProcessor(IPPROTO_TCP);
end;

Function TCP_Socket(Domain, SocketType, Protocol: Longint): Pointer;
Var
    PSock: PSocketRec;
begin
	{ Initialise new socket }
    New(PSock);
    PSock^.Status     := tcpClosed;
    PSock^.LocalIP    := 0;
    PSock^.LocalPort  := 0;
    PSock^.RemoteIP   := 0;
    PSock^.RemotePort := 0;
    PSock^.NextByteToSend := 1000;
    PSock^.LastByteConfirmed := 0;
    PSock^.LastByteReceived := 0;
    PSock^.DataBuffer := '';
    PSock^.PendingCount := 0;
    { Default Options }
    PSock^.ConnTimeout := 5000;
	{ Add new socket to the socket list }
    SockList^.Insert(PSock);
    { Done }
    TCP_Socket := PSock;
end;

Function TCP_Close(Sock: Pointer): LongInt;
Var
    Timeout: Comp;
	PSock: PSocketRec;
begin
	PSock := PSocketRec(Sock);
    { If the socket is connected, we disconnect it before proceeding }
    if PSock^.Status = tcpConnected then begin
        // Send close command
        PSock^.Status := tcpLocalEndWait1;
  	    Send(PSock, PSock^.LocalIP, PSock^.RemoteIP, PSock^.LocalPort,
          PSock^.RemotePort, CB_END or CB_ACK, nil, 0);
        // Wait for close to be confirmed
        Timeout := GetMSCounter + 5000;
        Repeat
            NET_Work_Low;
//      	WriteLn('Waiting for Socket to Close');
        until (GetMSCounter > Timeout) or (PSock^.Status = tcpClosed);
    end;
    { Free socket's data buffer }
    PSock^.DataBuffer := '';
    { Free the socket Structure, and remove it from the socket list }
    SockList^.Delete(PSock);
    Dispose(PSock);
    { Listo }
	TCP_Close := 0;
end;

Function TCP_ShutDown(Sock: Pointer; How: LongInt): LongInt;
Var
    Timeout: Comp;
	PSock: PSocketRec;
begin
	PSock := PSocketRec(Sock);
    { Check Socket status }
    if PSock^.Status <> tcpConnected then begin
        TCP_ShutDown := -1;
        LastError := WSAENOTCONN;
        Exit;
    end;
    // Send close command and wait for confirmation.
    Timeout := GetMSCounter + 5000;
    PSock^.Status := tcpLocalEndWait1;
    Send(PSock, PSock^.LocalIP, PSock^.RemoteIP, PSock^.LocalPort,
      PSock^.RemotePort, CB_END or CB_ACK, nil, 0);
    Repeat
        NET_Work_Low;
//      WriteLn('Waiting for Socket to Shut Down');
    until (GetMSCounter > Timeout) or (PSock^.Status in [tcpLocalEndWait2, tcpClosed]);
    { Done }
    if PSock^.Status in [tcpLocalEndWait2, tcpClosed] then begin
        TCP_ShutDown := 0;
    end else begin
        TCP_ShutDown := -1;
        LastError := WSAETIMEDOUT;
    end;
end;

Function TCP_Bind(Sock: Pointer; Var Addr; AddrLen: Longint): Boolean;
Var
	PSock: PSocketRec;
    InAddr: TInetSockAddr;
begin
    TCP_Bind := False;
    PSock := PSocketRec(Sock);
    InAddr := TInetSockAddr(Addr);
    { Check the given Address }
    if InAddr.family <> AF_INET then begin
        LastError := WSAEFAULT;
        Exit;
    end;
	{ If it's necessary, we choose a free local port for the socket }
    if InAddr.port = 0 then InAddr.port := GetFreePort;
    { Asignamos el IP y el puerto al Socket }
    PSock^.LocalIP := InAddr.addr;
    PSock^.LocalPort := InAddr.port;
    { Listo }
    TCP_Bind := True;
end;

Function TCP_Listen(Sock: Pointer; MaxConnect: Longint): Boolean;
Var
	PSock: PSocketRec;
begin
    PSock := PSocketRec(Sock);
    { If the socket is connected or in the process of getting connected, }
    { we cannot put it in passive mode, so we quit.                      }
    if PSock^.Status <> tcpClosed then begin
        TCP_Listen := False;
        LastError := WSAEISCONN;
		Exit;
    end;
    { If the socket is not bound to a valid Address and Port, we quit    }
    if (PSock^.LocalIP = INADDR_NONE) or (PSock^.LocalPort = 0) then begin
        TCP_Listen := False;
        LastError := WSAEINVAL;
	  	Exit;
    end;
    { Done }
    PSock^.Status := tcpListen;
    TCP_Listen := True;
end;

Function TCP_Accept(Sock: Pointer; Var Addr; Var Addrlen: Longint): Pointer;
Var
	i: Integer;
    Timeout: Comp;
    NewSock, PSock: PSocketRec;
begin
    PSock := PSocketRec(Sock);
    { If the socket is not listening, we quit }
    if PSock^.Status <> tcpListen then begin
        TCP_Accept := nil;
        LastError  := WSAEINVAL;
		Exit;
    end;
    { If there's not enough to copy the remote peer's address and port, }
    { we quit.                                                          }
    if AddrLen < SizeOf(TInetSockAddr) then begin
        TCP_Accept := nil;
        LastError := WSAEMSGSIZE;
        Exit;
    end;
    { If there's no incomming connection to accept, we quit }
    if PSock^.PendingCount <= 0 then begin
        TCP_Accept := nil;
        LastError  := WSAEWOULDBLOCK;
		Exit;
    end;
    { All Right. We create a new socket for the incomming connection }
    NewSock := TCP_Socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
    if NewSock = nil then begin
        TCP_Accept := nil;
        Exit;
    end;
    { The new socket has to be bound to the same local address and port }
    { as the listening socket. The remote IP and port of the new socket }
    { must be the ones from the connecting socket.                      }
    NewSock^.LocalIP 		  := PSock^.Pendings[1].LocalIP;
    NewSock^.RemoteIP         := PSock^.Pendings[1].RemoteIP;
    NewSock^.LocalPort        := PSock^.LocalPort;
    NewSock^.RemotePort       := PSock^.Pendings[1].RemotePort;
    NewSock^.LastByteReceived := PSock^.Pendings[1].SequenceNumber;
    { Change the status of the new socket  }
    NewSock^.Status := tcpSynRcvd;
    { Send the first confirmation to the remote end }
    Send(NewSock, PSock^.Pendings[1].LocalIP, PSock^.Pendings[1].RemoteIP,
      PSock^.LocalPort, PSock^.Pendings[1].RemotePort, CB_SYN or CB_ACK,
      nil, 0);
    { Wait for the connection to complete }
    Timeout := GetMSCounter + 5000;                     { 5 seconds }
    Repeat
        NET_Work_Low;
    Until (GetMSCounter > Timeout) or (NewSock^.Status in [tcpConnected, tcpClosed]);
    { If the connection was completed succesfully, we return the new  }
    { socket. Otherwise, we return nil, and the newly created socket  }
    { is destroyed.                                                   }
    if NewSock^.Status = tcpConnected then begin
        { Copy the remote Peer's address to the user-supplied buffer }
        NewSock^.NextByteToSend   := NewSock^.LastByteConfirmed;
        TCP_GetPeerName(NewSock, Addr, AddrLen);
        TCP_Accept := NewSock;
    end else begin
        TCP_Accept := nil;
        LastError := WSAETIMEDOUT;
        TCP_Close(NewSock);
    end;
    { Remove the Accepted connection from the pending queue }
    Dec(PSock^.PendingCount);
    for i := 1 to PSock^.PendingCount do begin
        PSock^.Pendings[i] := PSock^.Pendings[i + 1];
    end;
end;

Function TCP_Send(Sock: Pointer; Var Addr; AddrLen, Flags: Longint): Longint;
var
	P: PChar;
    Timeout: Comp;
    Tries: Integer;
    FragLen: DWORD;
    FragOk: Boolean;
    PSock: PSocketRec;
begin
    PSock := PSocketRec(Sock);

    { if the Socket is not connected, we quit }
    if PSock^.Status <> tcpConnected then begin
        TCP_Send  := SOCKET_ERROR;
        LastError := WSAENOTCONN;
		Exit;
    end;

    { All Right. Send the Data }
    P := @Addr;
    TCP_Send := AddrLen;
    While AddrLen > 0 do begin

		{ Flag to indicate that this fragment hasn't been sent yet }
        FragOk := False;

        { Calculate fragment size }
        if AddrLen > MTU then FragLen := MTU else FragLen := AddrLen;

        for Tries := 1 to 3 do begin

	        { Wait for the other end to be ready to receive one fragment }
    	    Timeout := GetMSCounter + 20000;
//	    	WriteLn(PSock^.RemoteWindow);
    	    While (GetMSCounter < Timeout) and (PSock^.RemoteWindow < FragLen)
        	  and (PSock^.Status = tcpConnected) do begin
//         		WriteLn('Waiting for Window to open');
	            NET_Work_Low;
    	    end;
        	if PSock^.RemoteWindow < FragLen then Break;

	        { Send the Fragment }
    	    if not Send(PSock, PSock^.LocalIP, PSock^.RemoteIP, PSock^.LocalPort,
        	  PSock^.RemotePort, CB_ACK, P, FragLen) then break;
	        Dec(PSock^.NextByteToSend, FragLen);
    	    if not Send(PSock, PSock^.LocalIP, PSock^.RemoteIP, PSock^.LocalPort,
	          PSock^.RemotePort, CB_ACK, P, FragLen) then break;

    	    { Wait for the other end to confirm the last fragment }
        	Timeout := GetMSCounter + 1000;
	        While (GetMSCounter < Timeout) and (PSock^.Status = tcpConnected)
    	      and (PSock^.NextByteToSend <> PSock^.LastByteConfirmed) do begin
        	    NET_Work_Low;
	        end;
            if (PSock^.LastByteConfirmed = PSock^.NextByteToSend) then begin
                FragOk := True;
                Break;
            end else begin
                Dec(PSock^.NextByteToSend, FragLen);
            end;

		end;

        { Go for next fragment. }
        if FragOk then begin
	        Dec(AddrLen, LongInt(FragLen));
            Inc(P, FragLen);
        end else begin
        	Break;
        end;
    end;

    // Check to see if everything could be sent. Otherwise, close the
    // socket and return an error condition.
    if AddrLen <> 0 then begin
        TCP_Send := SOCKET_ERROR;
        LastError := WSAECONNABORTED;
//      WriteLn('Socket cerrado por error en Send!');
        PSock^.Status := tcpClosed;
    end;
end;

Function TCP_Connect(Sock: Pointer; Var Addr; Addrlen: Longint): Boolean;
Var
    i: Integer;
    Timeout: Comp;
    Retries: Integer;
    PSock: PSocketRec;
    Gateway: TIPAddress;
    InAddr: TInetSockAddr;
begin
	PSock := PSocketRec(Sock);
    InAddr := TInetSockAddr(Addr);
    { Check the given Address }
    if InAddr.family <> AF_INET then begin
    	TCP_Connect := False;
        LastError := WSAEFAULT;
        Exit;
    end;
    { If the socket is already connected, we quit }
    if PSock^.Status <> tcpClosed then begin
        TCP_Connect := False;
        LastError := WSAEISCONN;
		Exit;
    end;
    { If the socket is not bound to a local IP, now it's the time to do it }
    if (PSock^.LocalIP = INADDR_ANY) or (PSock^.LocalIP = 0)
      or (PSock^.LocalIP = INADDR_BROADCAST) then begin
        if not IP_GetRouteForTarget(InAddr.addr, PSock^.LocalIP, Gateway) then begin
            TCP_Connect := False;
            LastError := WSAENETUNREACH;
        	Exit;
        end;
    end;
    { If the socket is not bound to a local port, now it's the time to do it }
    if PSock^.LocalPort = 0 then PSock^.LocalPort := GetFreePort;
    { Bind the socket to the remote peer }
    PSock^.RemoteIP := InAddr.addr;
    PSock^.RemotePort := InAddr.port;
    PSock^.NextByteToSend := 1000;
    PSock^.Status := tcpConnecting1;
    { The retries are given by the ConnTimeout parameter }
    Retries := PSock^.ConnTimeout div 500;
    if Retries < 1 then Retries := 1;
    for i := 1 to Retries do begin
        { Send a request for connection. }
        Send(PSock, PSock^.LocalIP, PSock^.RemoteIP, PSock^.LocalPort,
          PSock^.RemotePort, CB_SYN, nil, 0);
        { Wait for Ack }
        Timeout := GetMSCounter + 500;        { 1/2 second }
		Repeat
    	    NET_Work_Low;
        Until (GetMSCounter > Timeout) or (PSock^.Status in [tcpConnected, tcpClosed]);
        if PSock^.Status <> tcpConnecting1 then Break;
    end;
    { Did the connection end succesfully ? }
    case PSock^.Status of
		tcpConnected:
			begin
                PSock^.NextByteToSend := PSock^.LastByteConfirmed;
                TCP_Connect := True;
            end;
        tcpClosed:
        	begin
                TCP_Connect := False;
                LastError   := WSAECONNREFUSED;
			end;
        else
        	begin
                PSock^.Status := tcpClosed;
                TCP_Connect := False;
                LastError   := WSAETIMEDOUT;
            end;
    end;
end;

Function TCP_Recv(Sock: Pointer; Var Addr; AddrLen, Flags: Longint): Longint;
Var
    PSock: PSocketRec;
    AntWinSize: Integer;
begin
	PSock := PSocketRec(Sock);
    { If there's nothing to read, and the socket is not connected, we }
    { should return an error code. However, if there's something to   }
    { read, we should do it even if it's not connected.               }
    if (Length(PSock^.DataBuffer) = 0) then begin
        if not (PSock^.Status in [tcpConnected, tcpLocalEndWait1, tcpLocalEndWait2]) then begin
	        TCP_Recv  := SOCKET_ERROR;
            LastError := WSAENOTCONN;
        end else if AddrLen > 0 then begin
            TCP_Recv  := SOCKET_ERROR;
            LastError := WSAEWOULDBLOCK;
        end else begin
            TCP_Recv  := 0;
        end;
        Exit;
    end;
    { AddrLen = 0? }
    if AddrLen <= 0 then begin
        TCP_Recv := 0;
        Exit;    	
    end;
    { If there's not enough data to be read, we read only what we have }
    if AddrLen > Length(PSock^.DataBuffer) then
      AddrLen := Length(PSock^.DataBuffer);
    { Read data into the user-supplied buffer }
    Move(PSock^.DataBuffer[1], Addr, AddrLen);
    { If necessary, remove the fetched data from the buffer }
    if (Flags and MSG_PEEK) <> MSG_PEEK then begin
        PSock^.DataBuffer := Copy(PSock^.DataBuffer, AddrLen + 1,
          Length(PSock^.DataBuffer));
        { The reception window has changed, because more room is available }
        { now. Tell the other end.                                         }
        AntWinSize    := MAX_WINDOW - Length(PSock^.DataBuffer) - AddrLen;
        Send(PSock, PSock^.LocalIP, PSock^.RemoteIP,
          PSock^.LocalPort, PSock^.RemotePort, CB_ACK, nil, 0);
    end;
	TCP_Recv := AddrLen;
end;

Function TCP_ioctlsocket(Sock: Pointer; Option: LongInt; d: Pointer): LongInt;
Var
	dw: DWORD;
    PSock: PSocketRec;
begin
	NET_Work_Low;
    TCP_ioctlsocket := 0;
    PSock := PSocketRec(Sock);
    case Option of
        FIONREAD:
        	begin
                dw := Length(PSock^.DataBuffer);
                Move(dw, d^, 4);
            end;
        else
            begin
                TCP_ioctlsocket := SOCKET_ERROR;
                LastError := WSAEINVAL;
            end;
    end;
end;

Function TCP_GetSocketName  (Sock: Pointer; Var Addr; Var Addrlen: Longint): Longint;
Var
    PSock: PSocketRec;
begin
    PSock := PSocketRec(Sock);
    { Check to see if there's enough room in the destination buffer }
    if AddrLen < SizeOf(TInetSockAddr) then begin
        TCP_GetSocketName := SOCKET_ERROR;
        LastError := WSAEMSGSIZE;
    	Exit;
    end;
    { Copy the local IP and Port to the destination buffer }
    TInetSockAddr(Addr).family := AF_INET;
    TInetSockAddr(Addr).addr   := PSock^.LocalIP;
    TInetSockAddr(Addr).port   := PSock^.LocalPort;
    AddrLen := SizeOf(TInetSockAddr);
    { Done }
    TCP_GetSocketName := 0;
end;

Function TCP_GetPeerName(Sock: Pointer; Var Addr; Var Addrlen: Longint): Longint;
Var
    PSock: PSocketRec;
begin
    PSock := PSocketRec(Sock);
    { Check to see if there's enough room in the destination buffer }
    if AddrLen < SizeOf(TInetSockAddr) then begin
        TCP_GetPeerName := SOCKET_ERROR;
        LastError := WSAEMSGSIZE;
    	Exit;
    end;
    { Check to see if the socket is connected }
    if (PSock^.Status <> tcpConnected) then begin
        TCP_GetPeerName := SOCKET_ERROR;
        LastError := WSAENOTCONN;
    	Exit;
    end;
    { Copy the remote IP and Port to the destination buffer }
    TInetSockAddr(Addr).family := AF_INET;
    TInetSockAddr(Addr).addr   := PSock^.RemoteIP;
    TInetSockAddr(Addr).port   := PSock^.RemotePort;
    AddrLen := SizeOf(TInetSockAddr);
    { Done }
    TCP_GetPeerName := 0;
end;

Function TCP_SetSocketOptions(Sock: Pointer; Level, OptName: Longint; Var OptVal; OptLen: Longint): Longint;
Var
    PLong: ^LongInt;
    PSock: PSocketRec;
begin
    Result := SOCKET_ERROR;
    PSock  := PSocketRec(Sock);
    PLong  := @OptVal;
    if Level = SOL_SOCKET then begin
        case OptName of
            SOCKOPT_CONNTIMEOUT:
                if OptLen = SizeOf(LongInt) then begin
                    PSock^.ConnTimeout := PLong^;
                    Result := 0;
                end;
        end;
    end;
end;

Function TCP_GetSocketOptions(Sock: Pointer; Level, OptName: Longint; Var OptVal; Var OptLen: Longint): Longint;
Var
    PLong: ^LongInt;
    PSock: PSocketRec;
begin
    Result := SOCKET_ERROR;
    PSock  := PSocketRec(Sock);
    PLong  := @OptVal;
    if Level = SOL_SOCKET then begin
        case OptName of
            SOCKOPT_CONNTIMEOUT:
                if OptLen = SizeOf(LongInt) then begin
                    PLong^ := PSock^.ConnTimeout;
                    Result := 0;
                end;
        end;
    end;
end;



initialization
    LastError := 0;
    New(SockList, Init(10, 10));
finalization
    { If there's any unclosed socket, close it now }
    While SockList^.Count > 0 do TCP_Close(SockList^.At(0));
    { Free the socket collection }
    Dispose(SockList, Done);
end.
